/*
** Copyright (c) 1998, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished  -
** rights reserved under the Copyright Laws of the United States.
**
** File name:   d6stripk.c
**
** Description: Implementes DrawPrimitive strip drawing functions with
**              optimizations for 3DNow.
**
** Information: This file is included multiple times by d6strip.c.
**              It's used as a template for generating indexed and
**              non-indexed versions of DrawPrimitive API calls.
**
** $Revision: 9$
** $Date: 1/18/00 7:50:48 PM$
**
** $Log: 
**  9    3dfx      1.8         01/18/00 Bob Seitsinger  Changes to correctly write
**       to selected chip components (fbi,tmu0,tmu1). Chip component selection
**       occurred correctly for command packet header  writes, but needed to also
**       be done for data packet writes.
**  8    3dfx      1.7         01/13/00 Bob Seitsinger  Encase some variables in
**       #ifdef CMDFIFO. Also, fix some dereferncing syntax errors related to
**       structure element sOowfbi.
**  7    3dfx      1.6         01/10/00 Chris W. Shaw   Default to white diffuse in
**       the case that no diffuse is in FVF.  Fixes PRS 12055.
**  6    3dfx      1.5         12/14/99 Matt McClure    Fix for PRS 11794.  Added
**       WRAPn functionality and fixed autostripping bug with shadow maps on 3D
**       Winbench 2k Test 6 - Stations.
**  5    3dfx      1.4         12/06/99 Matt McClure    Mirror Russ Lind's fixes,
**       restored pF reference in code.
**  4    3dfx      1.3         12/06/99 Matt McClure    Fix typo in referencing
**       Flat shaded triangles.  Referencing pB when flat shading is wrong, should
**       be pA.  Fixes Flat shading quality test 1 in 3DWB 2k.
**  3    3dfx      1.2         12/01/99 Matt McClure    Fix for 3DMark 2000 lockup
**       in Test 2 Adventure, 3DMark 2000 sends down a FVF format that does not
**       contain a diffuse color, when accessing the entry in the FVF table for the
**       color calculation, the index of 128 was used and faulted on invalid memory
**       access.
**  2    3dfx      1.1         10/28/99 Christopher Wilcox Hardware definition
**       changes to merge divergent h3defs.h.
**  1    3dfx      1.0         09/11/99 StarTeam VTS Administrator 
** $
** 
** 10    9/02/99 11:15a Cwilcox
** Added 24bpp zbuffer support to Direct3D.
** 
** 9     8/17/99 10:46a Msmith
** I added include statements for stbperf.inc so we can enable the
** C_AUTOSTRIP code.
** 
** 8     8/09/99 9:56a Msmith
** Added femms around some 3dnow code for DCT fix. Bob J.
** 
** 7     8/06/99 2:47p Msmith
** More DCT and auto-strip changes from Bob J.
** 
** 6     8/04/99 2:22p Msmith
** fixes for DCT tests from Bob J.
** 
** 5     7/28/99 11:59p Bseitsin
** Changes to enable DX7 for W9x.
** 
** 4     7/13/99 2:28p Cshaw
** Added runtime support for Napalm's multitexturing (NAPALM_CU).
** 
** 3     6/04/99 10:34a Cshaw
** Updated code for 2pass specular with combineMode.
** 
** 2     6/03/99 3:21p Cshaw
** Added combineMode for 2 pass specular.
** 
** 1     6/02/99 6:45a Michael
** Branch from H3
** 
** 7     5/24/99 5:06p Bseitsin
** Removal of Antialiasing code.
** 
** 6     4/09/99 12:35p Stb_bseitsin
** Added Napalm registers. Added ifdef H5.
** 
** 5     1/25/99 4:53p Peterm
** added unified header information
** 
** 4     12/16/98 7:20p Sreid
** Latest K6 drop from contractor
**
** 2     11/05/98 12:07p Adrians
** Remove a compiler warning generated by the K6-2 optimisations.
**
** 1     11/04/98 4:10p Adrians
** Updated changes to K6-2.
**
** 1     10/15/98 4:13p stit (Metabyte Inc)
** Created
*/

// include #defines for performance optimizations
#ifdef INCSTBPERF
#include "..\build\stbperf.inc"
#endif


#ifdef C_AUTOSTRIP
// RAJ 8/4/99 Defines used by the 'C' autostrip logic
#define BIT22  (1<<22)
#define BIT23  (1<<23)
#define NO_CONTINUE  (1<<0)
#define FORCE_SINGLE (1<<1)
#else
#pragma message("Not compiling with C_AUTOSTRIP optimizations ")
#endif

#ifdef K6_2

#if ( DX >= 6 )

#pragma warning( disable: 4799)

#include "3DNow_VC.h"

#define ALIGN8 1

/*-------------------------------------------------------------------
Function Name:  dp2StripAll_K62O

Description:    This function handles drawing DrawPrimitive strips in
                solid-filled mode.  It's optimized via inline
                assembly for 3DNow.

Return:         void
-------------------------------------------------------------------*/

void dp2StripAll_K62O( RC *pRc, DWORD count, LPWORD idx, LPDWORD vertices, DWORD vertexType )
{
  LPDWORD       pA, pB, pC, pF;
#ifdef C_AUTOSTRIP
  //RAJ 8/4/99 Locals used by 'C' Auto Striping
  LPDWORD       PrevA, PrevB, PrevC;
  FxU32         DummyPH;
  FxU32         LastTriType = BIT22;
  DWORD         DoInitialTriangle = NO_CONTINUE;
#ifdef CMDFIFO
  DWORD         ContinueTest;
  DWORD         ContinueTemp;
#endif
#endif
  float         s1, s2, s3, t1, t2, t3;
  float         w1, w2, w3;
#ifdef DCT_FIX
  float         wb1, wb2, wb3;
#endif
  float         z1, z2, z3;
#if (NUMTEXTUREUNITS > 1)
  float         s1a, t1a, s1b, t1b, s1c, t1c;
#endif
  D3DCOLOR      aColor, bColor, cColor;
  int           sign;
  FxU32         setupFlag = pRc->sst.sSetupMode;
  BOOL          flip = 0;
#if (!(NUMTEXTUREUNITS > 1))
  int ST0_W0_Z_RGB_A =
    (int)((setupFlag & 0x3f) == (SST_SETUP_RGB | SST_SETUP_A | SST_SETUP_Z | SST_SETUP_W0 | SST_SETUP_ST0));
#endif
  DWORD offs_SX, offs_SY, offs_SZ, offs_RHW, pix_offs;
  __int64 tmp_pa, tmp_pb, tmp_pc ;

  SETUP_PPDEV(pRc)

  CMDFIFO_PROLOG(cmdFifo);

  __asm{ femms }

  offs_SX=FVFO_SX; offs_SY=FVFO_SY;
  offs_SZ=FVFO_SZ; offs_RHW=FVFO_RHW;
   pix_offs=*((int*)&(PIXEL_OFFSET));

#ifdef C_AUTOSTRIP
  PrevA = PrevB = PrevC = 0;
#endif

  TRI_INIT;

  for( ; count > 0; count-- )
  {
    TRI_NEXT;

    //((float*)&sign)[0] = ((FLTP(pA)[FVFO_SX] - FLTP(pB)[FVFO_SX]) * (FLTP(pB)[FVFO_SY] - FLTP(pC)[FVFO_SY])) -
    //                     ((FLTP(pB)[FVFO_SX] - FLTP(pC)[FVFO_SX]) * (FLTP(pA)[FVFO_SY] - FLTP(pB)[FVFO_SY]));
    __asm {
                mov     edx, offs_SX
                mov     eax, pA

                shl     edx, 2
                mov     ecx, pB

                movd    mm0, [eax+edx]   //undef|pA_SX           in mm0
                movd    mm1, [ecx+edx]   //undef|pB_SX           in mm1

                mov     eax, pC
                punpckldq mm0, mm1       //pB_SX|pA_SX           in mm0

                punpckldq mm1, [eax+edx] //pC_SX|pB_SX           in mm1
                mov     edx, offs_SY

                pfsub(mm0_mm1)           //pB_SX-pC_SX|pA_SX-pB_SX in mm0
                shl     edx, 2

                movd    mm2, [ecx+edx]   //undef|pB_SY           in mm2
                mov     ecx, pA

                movd    mm3, [eax+edx]   //undef|pC_SY           in mm3
                punpckldq mm2, [ecx+edx] //pA_SY|pB_SY           in mm2

                punpckldq mm3, mm2       //pB_SY|pC_SY           in mm3
                pfsub(mm2_mm3)

                pfmul(mm0_mm2)
                movq    mm1, mm0

                punpckhdq mm0, mm0
                pfsub(mm1_mm0)

                movd    dword ptr sign, mm1
        }

    sign &= 0x80000000;

    if ((sign ^ pRc->cullMask) != 0x80000000)
    {
    #if (NUMTEXTUREUNITS > 1)
      CMDFIFO_CHECKROOM( cmdFifo, (PH3_SIZE * 2) + 30 + 15  + (PH1_SIZE * 7) + 8 );
    #else
      CMDFIFO_CHECKROOM( cmdFifo, (PH3_SIZE * 2) + 24 + 15  + (PH1_SIZE * 7) + 8 + ALIGN8);
    #endif

      /* 3DMark 2000 does not send down a diffuse color in their FVF Format */
	  /* Trying to index the 128th entry in pA, pB, or pC will fault        */
      if ( !(FVFO_COLOR > FVFO_SIZE) )
	  {
        if ( pRc->shadeMode == D3DSHADE_FLAT )
        {
#ifdef C_AUTOSTRIP
          // HW cannot support FLAT_SHADE and Auto Strip/Fan
          // Force SINGLE TRIANGLES
          DoInitialTriangle  |= FORCE_SINGLE;
#endif

          if ( (pRc->specular) && (pRc->texture == 0) )
            CLAMP888( aColor, pF[FVFO_COLOR], pF[FVFO_SPECULAR] );
          else
            aColor = pF[FVFO_COLOR];

          bColor = aColor;
          cColor = aColor;
        }
        else
        {
          if ( (pRc->specular) && (pRc->texture == 0) )
          {
            CLAMP888(aColor, pA[FVFO_COLOR], pA[FVFO_SPECULAR]);
            CLAMP888(bColor, pB[FVFO_COLOR], pB[FVFO_SPECULAR]);
            CLAMP888(cColor, pC[FVFO_COLOR], pC[FVFO_SPECULAR]);
          }
          else
          {
            aColor = pA[FVFO_COLOR];
            bColor = pB[FVFO_COLOR];
            cColor = pC[FVFO_COLOR];
          }
        }
	  }
	  else /* Diffuse Color not in FVF Format. default to white*/
	  {
        aColor = 0xffffffff;
        bColor = 0xffffffff;
        cColor = 0xffffffff;
      }

#ifdef DCT_FIX
      __asm{ femms }
      //RAJ 8/4/99 Needs to be put down in 3DNow! assembly
      w1 = FLTP(pA)[FVFO_RHW] * pRc->scaleW;
      w2 = FLTP(pB)[FVFO_RHW] * pRc->scaleW;
      w3 = FLTP(pC)[FVFO_RHW] * pRc->scaleW;
      __asm{ femms }
#endif

      if( pRc->state & STATE_REQUIRES_WBUFFER ) {
#ifdef DCT_FIX
        //wb1 = w1;
        //wb2 = w2;
        //wb3 = w3;
        *((int*)&wb1) = *((int*)&w1);
        *((int*)&wb2) = *((int*)&w2);
        *((int*)&wb3) = *((int*)&w3);
#else
        //w1 = WSCALE( FLTP(pA)[FVFO_RHW] );
        //w2 = WSCALE( FLTP(pB)[FVFO_RHW] );
        //w3 = WSCALE( FLTP(pC)[FVFO_RHW] );
        __asm {
                mov     eax, pRc
                movd    mm3, [eax]RC.aW         //undef|pRc->aW in mm3
                mov     edx, offs_RHW
                movd    mm4, [eax]RC.bW         //undef|pRc->bW in mm4
                shl     edx, 2
                mov     ecx, pA
                movd    mm0, [ecx+edx]       //undef|pA_RHW
                mov     ecx, pB
                pfmul(mm0_mm3)
                movd    mm1, [ecx+edx]       //undef|pA_RHW
                mov     ecx, pC
                pfadd(mm0_mm4)
                pfmul(mm1_mm3)
                movd    mm2, [ecx+edx]       //undef|pA_RHW
                movd    w1, mm0
                pfadd(mm1_mm4)
                pfmul(mm2_mm3)
                movd    w2, mm1
                pfadd(mm2_mm4)
                movd    w3, mm2
        }
#endif
        if (pRc->state & STATE_REQUIRES_VERTEXFOG)
        {
          //z1 = (float)((255 - RGBA_GETALPHA(pA[FVFO_SPECULAR])) << 8);
          //z2 = (float)((255 - RGBA_GETALPHA(pB[FVFO_SPECULAR])) << 8);
          //z3 = (float)((255 - RGBA_GETALPHA(pC[FVFO_SPECULAR])) << 8);
          __int64 iqz1;
          int iz2;

          *((int*)&iqz1)    = (255 - RGBA_GETALPHA(pA[FVFO_SPECULAR])) << 8;
          *((int*)&iqz1+1)  = (255 - RGBA_GETALPHA(pB[FVFO_SPECULAR])) << 8;
          iz2 = (255 - RGBA_GETALPHA(pC[FVFO_SPECULAR])) << 8;
          __asm {
                movq    mm0, iqz1
                movd    mm1, iz2
                pi2fd(mm0_mm0)
                pi2fd(mm1_mm1)
                movd    z1, mm0
                punpckhdq mm0, mm0
                movd    z3, mm1
                movd    z2, mm0
          }
        }
      }
      else
      {
        //z1 = ZSCALE( FLTP(pA)[FVFO_SZ] );
        //z2 = ZSCALE( FLTP(pB)[FVFO_SZ] );
        //z3 = ZSCALE( FLTP(pC)[FVFO_SZ] );
        const float f_zscale = pRc->zScale;
        __asm{
                mov     edx, offs_SZ
                mov     ecx, pA

                shl     edx, 2
                movd    mm3, f_zscale

                movd    mm0, [ecx+edx]       //undef|pA_SZ
                mov     ecx, pB

                movd    mm1, [ecx+edx]       //undef|pB_SZ
                pfmul(mm0_mm3)

                mov     ecx, pC
                movd    mm2, [ecx+edx]       //undef|pC_SZ

                pfmul(mm1_mm3)
                movd    z1, mm0

                pfmul(mm2_mm3)
                movd    z2, mm1

                movd    z3, mm2
        }


        if (pRc->state & STATE_REQUIRES_VERTEXFOG)
        {
#ifdef DCT_FIX
          //wb1 = (float)(255 - RGBA_GETALPHA(pA[FVFO_SPECULAR]));
          //wb2 = (float)(255 - RGBA_GETALPHA(pB[FVFO_SPECULAR]));
          //wb3 = (float)(255 - RGBA_GETALPHA(pC[FVFO_SPECULAR]));
          __int64 iqw1;
          int iw2;
          *((int*)&iqw1)    = 255 - RGBA_GETALPHA(pA[FVFO_SPECULAR]);
          *((int*)&iqw1+1)  = 255 - RGBA_GETALPHA(pB[FVFO_SPECULAR]);
          iw2               = 255 - RGBA_GETALPHA(pC[FVFO_SPECULAR]);
          __asm {
                movq    mm0, iqw1
                movd    mm1, iw2
                pi2fd(mm0_mm0)
                pi2fd(mm1_mm1)
                movd    wb1, mm0
                punpckhdq mm0, mm0
                movd    wb3, mm1
                movd    wb2, mm0
          }
#else
          //w1 = (float)(255 - RGBA_GETALPHA(pA[FVFO_SPECULAR]));
          //w2 = (float)(255 - RGBA_GETALPHA(pB[FVFO_SPECULAR]));
          //w3 = (float)(255 - RGBA_GETALPHA(pC[FVFO_SPECULAR]));
          __int64 iqw1;
          int iw2;
          *((int*)&iqw1)    = 255 - RGBA_GETALPHA(pA[FVFO_SPECULAR]);
          *((int*)&iqw1+1)  = 255 - RGBA_GETALPHA(pB[FVFO_SPECULAR]);
          iw2               = 255 - RGBA_GETALPHA(pC[FVFO_SPECULAR]);
          __asm {
                movq    mm0, iqw1
                movd    mm1, iw2
                pi2fd(mm0_mm0)
                pi2fd(mm1_mm1)
                movd    w1, mm0
                punpckhdq mm0, mm0
                movd    w3, mm1
                movd    w2, mm0
          }
#endif
        }
        else if (pRc->state & STATE_REQUIRES_HWFOG)
        {
#ifdef DCT_TEST
          //wb1 = FLTP(pA)[FVFO_RHW];
          //wb2 = FLTP(pB)[FVFO_RHW];
          //wb3 = FLTP(pC)[FVFO_RHW];
          *((int*)&wb1) = ((int*)pA)[FVFO_RHW];
          *((int*)&wb2) = ((int*)pB)[FVFO_RHW];
          *((int*)&wb3) = ((int*)pC)[FVFO_RHW];
#else
          //w1 = FLTP(pA)[FVFO_RHW];
          //w2 = FLTP(pB)[FVFO_RHW];
          //w3 = FLTP(pC)[FVFO_RHW];
          *((int*)&w1) = ((int*)pA)[FVFO_RHW];
          *((int*)&w2) = ((int*)pB)[FVFO_RHW];
          *((int*)&w3) = ((int*)pC)[FVFO_RHW];
#endif
        }
      }

      // First texture stage
      if( pRc->state & STATE_REQUIRES_ST_TMU0 )
      {
        float lscaleS, lscaleT;
        float txl_soffset, txl_toffset;

        //s1 = FLTP(pA)[FVFO_TU + pRc->t0CoordIndex];
        //t1 = FLTP(pA)[FVFO_TV + pRc->t0CoordIndex];
        //s2 = FLTP(pB)[FVFO_TU + pRc->t0CoordIndex];
        //t2 = FLTP(pB)[FVFO_TV + pRc->t0CoordIndex];
        //s3 = FLTP(pC)[FVFO_TU + pRc->t0CoordIndex];
        //t3 = FLTP(pC)[FVFO_TV + pRc->t0CoordIndex];
        //lscaleS = pRc->sst.scaleS;
        //lscaleT = pRc->sst.scaleT;

        *((int*)&s1) = ((int*)pA)[FVFO_TU + pRc->t0CoordIndex];
        *((int*)&t1) = ((int*)pA)[FVFO_TV + pRc->t0CoordIndex];

        *((int*)&s2) = ((int*)pB)[FVFO_TU + pRc->t0CoordIndex];
        *((int*)&t2) = ((int*)pB)[FVFO_TV + pRc->t0CoordIndex];

        *((int*)&s3) = ((int*)pC)[FVFO_TU + pRc->t0CoordIndex];
        *((int*)&t3) = ((int*)pC)[FVFO_TV + pRc->t0CoordIndex];

        *((int*)&lscaleS) = *((int*)&(pRc->sst.scaleS));
        *((int*)&lscaleT) = *((int*)&(pRc->sst.scaleT));

        *((int*)&txl_soffset) = *((int*)&(TEXEL_SOFFSET));
        *((int*)&txl_toffset) = *((int*)&(TEXEL_TOFFSET));

#ifdef C_AUTOSTRIP
        if ( (pRc->wrapT0 & D3DWRAP_U) || (pRc->wrapT0 & D3DWRAP_V))
		{
		  DoInitialTriangle |= FORCE_SINGLE;
		}
#endif

		// two ways to texture. D3D wraps its textures going around the other direction
        // so we need to adjust S and T in order to get the correct result

        //WRAP( s, (pRc->wrapT0 & D3DWRAP_U) );
        //WRAP( t, (pRc->wrapT0 & D3DWRAP_V) );
        WRAP_3DNow( s, (pRc->wrapT0 & D3DWRAP_U) );
        WRAP_3DNow( t, (pRc->wrapT0 & D3DWRAP_V) );

        // compute the slope of s t and w
        // NOTE: D3D passes in S, T and 1/W and we need S/W, T/W and 1/W
        if( pRc->state & STATE_REQUIRES_PERSPECTIVE )
        {
#ifdef DCT_FIX
          __asm{ femms }
          //  !!! Needs to be put down into 3DNow ASM !!!!
          s1 = ( ( s1 * lscaleS ) + TEXEL_SOFFSET ) * w1;
          s2 = ( ( s2 * lscaleS ) + TEXEL_SOFFSET ) * w2;
          s3 = ( ( s3 * lscaleS ) + TEXEL_SOFFSET ) * w3;

          t1 = ( ( t1 * lscaleT ) + TEXEL_TOFFSET ) * w1;
          t2 = ( ( t2 * lscaleT ) + TEXEL_TOFFSET ) * w2;
          t3 = ( ( t3 * lscaleT ) + TEXEL_TOFFSET ) * w3;
          __asm{ femms }
#else
          //s1 = ((s1 * lscaleS) + TEXEL_SOFFSET) * FLTP(pA)[FVFO_RHW];
          //s2 = ((s2 * lscaleS) + TEXEL_SOFFSET) * FLTP(pB)[FVFO_RHW];
          //s3 = ((s3 * lscaleS) + TEXEL_SOFFSET) * FLTP(pC)[FVFO_RHW];

          //t1 = ((t1 * lscaleT) + TEXEL_TOFFSET) * FLTP(pA)[FVFO_RHW];
          //t2 = ((t2 * lscaleT) + TEXEL_TOFFSET) * FLTP(pB)[FVFO_RHW];
          //t3 = ((t3 * lscaleT) + TEXEL_TOFFSET) * FLTP(pC)[FVFO_RHW];
          __asm {
                movd    mm0, s1
                movd    mm3, lscaleS

                punpckldq mm0, s2              //      s2|s1   in mm0
                movd    mm4, txl_soffset

                punpckldq mm3, mm3             //lscaleS|lscaleS       in mm3
                mov     eax, pA

                pfmul(mm0_mm3)                        //s2 * lscaleS|s1 * lscaleS
                mov     edx, offs_RHW

                punpckldq mm4, mm4
                movd    mm7, [eax+edx*4]             //|pA[FVFO_RHW]

                pfadd(mm0_mm4)                        //s2|s1 in mm0
                mov     eax, pB

                punpckldq mm7, [eax+edx*4]             //pB[FVFO_RHW]|pA[FVFO_RHW]
                movd    mm1, t1

                movd    mm5, lscaleT
                pfmul(mm0_mm7)

                punpckldq mm1, t2              //      t2|t1   in mm1
                punpckldq mm5, mm5

                movd    s1, mm0
                movd    mm6, txl_toffset

                pfmul(mm1_mm5)
                movd    mm2, s3

                punpckhdq mm0, mm0
                punpckldq mm6, mm6

                movd    s2, mm0
                punpckldq mm2, t3              //      t3|s3   in mm2

                pfadd(mm1_mm6)
                mov     eax, pC

                punpckldq mm3, mm5             //      lscaleT|lscaleS in mm3
                movd    mm0, [eax+edx*4]             //|pC[FVFO_RHW]

                pfmul(mm1_mm7)
                punpckldq mm4, mm6             //TEXEL_TOFFSET|TEXEL_SOFFSET   in mm4

                pfmul(mm2_mm3)
                punpckldq mm0, mm0                             //pC[FVFO_RHW]|pC[FVFO_RHW]

                movd    t1, mm1
                pfadd(mm2_mm4)

                punpckhdq mm1, mm1
                pfmul(mm2_mm0)

                movd    t2, mm1
                movd    s3, mm2

                punpckhdq mm2, mm2
                movd      t3, mm2
          }
#endif  //DCT_FIX
        }
        // if the texture is not prespective correct then the w is effectively equal to 1
        else
        {
          //s1 = (s1 * lscaleS) + TEXEL_SOFFSET;
          //s2 = (s2 * lscaleS) + TEXEL_SOFFSET;
          //s3 = (s3 * lscaleS) + TEXEL_SOFFSET;

          //t1 = (t1 * lscaleT) + TEXEL_TOFFSET;
          //t2 = (t2 * lscaleT) + TEXEL_TOFFSET;
          //t3 = (t3 * lscaleT) + TEXEL_TOFFSET;
          __asm {
                movd    mm0, s1
                movd    mm3, lscaleS

                punpckldq mm0, s2              //      s2|s1   in mm0
                movd    mm4, txl_soffset

                punpckldq mm3, mm3             //lscaleS|lscaleS       in mm3
                movd    mm1, t1

                pfmul(mm0_mm3)                        //s2 * lscaleS|s1 * lscaleS
                movd    mm5, lscaleT

                punpckldq mm4, mm4
                punpckldq mm1, t2              //      t2|t1   in mm1

                pfadd(mm0_mm4)                        //s2|s1 in mm0
                movd    mm6, txl_toffset

                punpckldq mm5, mm5
                movd    mm2, s3

                pfmul(mm1_mm5)
                movd    s1, mm0

                punpckldq mm6, mm6
                punpckldq mm2, t3              //      t3|s3   in mm2

                pfadd(mm1_mm6)
                punpckldq mm3, mm5             //      lscaleT|lscaleS in mm3

                movd    t1, mm1
                punpckhdq mm0, mm0

                pfmul(mm2_mm3)
                punpckldq mm4, mm6             //TEXEL_TOFFSET|TEXEL_SOFFSET   in mm4

                movd    s2, mm0
                punpckhdq mm1, mm1

                pfadd(mm2_mm4)
                movd    t2, mm1

                movd    s3, mm2
                punpckhdq mm2, mm2

                movd    t3, mm2
          }
        }
#ifdef C_AUTOSTRIP
        // Check for continuation WRAP bug
        // If any s or t value is greater than
        // 256.0, bail out of Auto Striping
        // for the remainder of the primitive
        // !! This should be put down into 3DNow! assembler !!
        __asm{ femms }
        if((s1 > 257.0) || (s2 > 257.0) || (s3 > 257.0) ||
           (t1 > 257.0) || (t2 > 257.0) || (t3 > 257.0))
        {
           DoInitialTriangle  |= FORCE_SINGLE;
        }
        __asm{ femms }
#endif
      } // texture

    #if (NUMTEXTUREUNITS > 1)
      // Second texture stage
      if( pRc->state & STATE_REQUIRES_ST_TMU1 ) {
        float lscaleS, lscaleT;

        __asm{ femms }

        s1a = FLTP(pA)[FVFO_TU + pRc->t1CoordIndex];
        t1a = FLTP(pA)[FVFO_TV + pRc->t1CoordIndex];

        s1b = FLTP(pB)[FVFO_TU + pRc->t1CoordIndex];
        t1b = FLTP(pB)[FVFO_TV + pRc->t1CoordIndex];

        s1c = FLTP(pC)[FVFO_TU + pRc->t1CoordIndex];
        t1c = FLTP(pC)[FVFO_TV + pRc->t1CoordIndex];

        lscaleS = pRc->sst.scaleS1;
        lscaleT = pRc->sst.scaleT1;

#ifdef C_AUTOSTRIP
        if ( (pRc->wrapT1 & D3DWRAP_U) || (pRc->wrapT1 & D3DWRAP_V))
		{
		  DoInitialTriangle |= FORCE_SINGLE;
		}
#endif

        // two ways to texture. D3D wraps its textures going around the other direction
        // so we need to adjust S and T in order to get the correct result
        DX6_WRAP( s1, (pRc->wrapT1 & D3DWRAP_U) );
        DX6_WRAP( t1, (pRc->wrapT1 & D3DWRAP_V) );

        // compute the slope of s t and w
        // NOTE: D3D passes in S, T and 1/W and we need S/W, T/W and 1/W
        if( pRc->state & STATE_REQUIRES_PERSPECTIVE )
        {
#ifdef DCT_FIX
          s1a = ( ( s1a * lscaleS ) + TEXEL_S1OFFSET ) * w1;
          s1b = ( ( s1b * lscaleS ) + TEXEL_S1OFFSET ) * w2;
          s1c = ( ( s1c * lscaleS ) + TEXEL_S1OFFSET ) * w3;

          t1a = ( ( t1a * lscaleT ) + TEXEL_T1OFFSET ) * w1;
          t1b = ( ( t1b * lscaleT ) + TEXEL_T1OFFSET ) * w2;
          t1c = ( ( t1c * lscaleT ) + TEXEL_T1OFFSET ) * w3;
#else
          s1a = ( ( s1a * lscaleS ) + TEXEL_S1OFFSET ) * FLTP(pA)[FVFO_RHW];
          s1b = ( ( s1b * lscaleS ) + TEXEL_S1OFFSET ) * FLTP(pB)[FVFO_RHW];
          s1c = ( ( s1c * lscaleS ) + TEXEL_S1OFFSET ) * FLTP(pC)[FVFO_RHW];

          t1a = ( ( t1a * lscaleT ) + TEXEL_T1OFFSET ) * FLTP(pA)[FVFO_RHW];
          t1b = ( ( t1b * lscaleT ) + TEXEL_T1OFFSET ) * FLTP(pB)[FVFO_RHW];
          t1c = ( ( t1c * lscaleT ) + TEXEL_T1OFFSET ) * FLTP(pC)[FVFO_RHW];
#endif
        }
        // if the texture is not prespective correct then the w is effectively equal to 1
        else
        {
          s1a = ( s1a * lscaleS ) + TEXEL_S1OFFSET;
          s1b = ( s1b * lscaleS ) + TEXEL_S1OFFSET;
          s1c = ( s1c * lscaleS ) + TEXEL_S1OFFSET;

          t1a = ( t1a * lscaleT ) + TEXEL_T1OFFSET;
          t1b = ( t1b * lscaleT ) + TEXEL_T1OFFSET;
          t1c = ( t1c * lscaleT ) + TEXEL_T1OFFSET;
        }
#ifdef C_AUTOSTRIP
        // Check for continuation WRAP bug
        // If any s or t value is greater than
        // 256.0, bail out of Auto Striping
        // for the remainder of the primitive
        if((s1a > 257.0) || (s1b > 257.0) || (s1c > 257.0) ||
           (t1a > 257.0) || (t1b > 257.0) || (t1c > 257.0))
        {
           DoInitialTriangle  |= FORCE_SINGLE;
        }
#endif

      } // texture
      __asm{ femms }
    #endif
#if  defined( CMDFIFO ) && defined( C_AUTOSTRIP )
      //****************************
      // Auto Strip Logic
      // Test for Auto Strip Case
      //****************************
      if((PrevC == pA) && (PrevB == pB))
      {
         // Strip Found!
         DoInitialTriangle &= ~NO_CONTINUE;
         ContinueTest = BIT23;          // For Strip!
         ContinueTemp = ContinueTest;
         ContinueTest ^= LastTriType;
         LastTriType = ContinueTemp;
         ContinueTest ^= BIT22;
         ContinueTest &= BIT22;
      }
      else if((PrevA == pA) && (PrevC == pB)) 
      {
         // Fan Found!
         DoInitialTriangle &= ~NO_CONTINUE;
         ContinueTest = BIT22;          // For Fan!
         ContinueTemp = ContinueTest;
         ContinueTest ^= LastTriType;
         LastTriType = ContinueTemp;
         ContinueTest ^= BIT22;
         ContinueTest &= BIT22;
      }
      else
      {
         //No connection invloved
         DoInitialTriangle |= NO_CONTINUE;
         LastTriType = BIT22;
      }

      PrevA = pA;
      PrevB = pB;
      PrevC = pC;

      if(DoInitialTriangle) 
      {
#endif

#if (!(NUMTEXTUREUNITS > 1))
    if (ST0_W0_Z_RGB_A) {
      FxU32 *mycmdFifo;

      if (!((int)cmdFifo & 4)) SETPH( cmdFifo, 0 );
      SETPH( cmdFifo, CMDFIFO_BUILD_PK3( CMD_TRI, 3, setupFlag, 1 ) );

      //double q_val1,q_val2,q_val3,q_val4;
      //double q_val5,q_val6,q_val7,q_val8;
      //double q_val9,q_val10;

      mycmdFifo = cmdFifo + hwIndex;
      //Send 1st vertex
      //*((float*)&q_val1)       = FLTP(pA)[FVFO_SX] + PIXEL_OFFSET ;
      //*((float*)&q_val1+1)     = FLTP(pA)[FVFO_SY] + PIXEL_OFFSET ;
      //*(float*)&q_val2         = *((float*)&aColor) ;
      //*((float*)&q_val2+1)     = z1 ;
      //*(float*)&q_val3         = FLTP(pA)[FVFO_RHW] ;
      //*((float*)&q_val3+1)     = s1 ;
      //*((float*)&q_val4)       = t1;
      //*((float*)&q_val4+1)     = FLTP(pB)[FVFO_SX] + PIXEL_OFFSET ;

      // Send 2nd vertex
      //*(float*)&q_val5         = FLTP(pB)[FVFO_SY] + PIXEL_OFFSET ;
      //*((float*)&q_val5+1)     = *(float*)&bColor ;
      //*((float*)&q_val6)       = z2 ;
      //*((float*)&q_val6+1)     = FLTP(pB)[FVFO_RHW] ;
      //*((float*)&q_val7)       = s2 ;
      //*((float*)&q_val7+1)     = t2 ;

      // Send 3rd vertex
      //*(float*)&q_val8         = FLTP(pC)[FVFO_SX] + PIXEL_OFFSET ;
      //*((float*)&q_val8+1)     = FLTP(pC)[FVFO_SY] + PIXEL_OFFSET ;
      //*(float*)&q_val9         = *(float*)&cColor ;
      //*((float*)&q_val9+1)     = z3 ;
      //*(float*)&q_val10        = FLTP(pC)[FVFO_RHW] ;
      //*((float*)&q_val10+1)    = s3 ;
      __asm {
                mov     ecx, pA
                movd    mm7, pix_offs

                mov     edx, offs_SX
                movd    mm0, [ecx+edx*4]

                punpckldq mm7, mm7
                mov     edx, offs_SY

                punpckldq mm0, [ecx+edx*4]             //pA[FVFO_SY]|pA[FVFO_SX]
                mov     eax, mycmdFifo

                movd    mm1, aColor
                pfadd(mm0_mm7)

                mov     edx, offs_RHW
                punpckldq mm1, z1

                movd    mm2, [ecx+edx*4]
                movq    [eax   ], mm0

                punpckldq mm2, s1
                mov     ecx, pB

                movq    [eax+8 ], mm1
                movd    mm3, t1

                mov     edx, offs_SY
                movq    [eax+16 ], mm2

                movd    mm4, [ecx+edx*4]             //|pB[FVFO_SY]  in mm4
                mov     edx, offs_SX

                punpckldq mm4, [ecx+edx*4]             //pB[FVFO_SX]|pB[FVFO_SY]       in mm4
                movd    mm5, z2

                pfadd(mm4_mm7)
                mov     edx, offs_RHW

                movq    mm0, mm4
                punpckldq mm4, bColor

                punpckhdq mm0, mm0
                punpckldq mm5, [ecx+edx*4]

                punpckldq mm3, mm0
                movd    mm6, s2

                mov     ecx, pC
                movq    [eax+24], mm3

                punpckldq mm6, t2
                movq    [eax+32], mm4

                movd    mm3, [ecx+edx*4]             //|pC[FVFO_RHW]
                mov     edx, offs_SX

                movq    [eax+40], mm5
                movd    mm1, [ecx+edx*4]

                mov     edx, offs_SY
                movq    [eax+48], mm6

                punpckldq mm1, [ecx+edx*4]             //pC[FVFO_SY]|pC[FVFO_SX]
                movd    mm2, cColor

                pfadd(mm1_mm7)
                punpckldq mm2,    z3

                movq    [eax+56], mm1
                punpckldq mm3, s3

                movq    [eax+64], mm2
                movq    [eax+72], mm3
          }
          *((int* )mycmdFifo + 6 +7+ 7) = *(int*)&t3;
          hwIndex += 21;
    }
    else
#endif  //end #if (!(NUMTEXTUREUNITS > 1))
      {
     #ifdef C_AUTOSTRIP
        DummyPH = (CMDFIFO_BUILD_PK3( CMD_START, 3, setupFlag, 1 ) | BIT22);
        SETPH( cmdFifo, DummyPH);
     #else
        SETPH( cmdFifo, CMDFIFO_BUILD_PK3( CMD_TRI, 3, setupFlag, 1 ) );
     #endif

        //Send 1st vertex
        __asm {
                mov     ecx, pA
                movd    mm7, pix_offs

                mov     edx, offs_SX
                movd    mm0, [ecx+edx*4]

                punpckldq mm7, mm7
                mov     edx, offs_SY

                punpckldq mm0, [ecx+edx*4]             //pA[FVFO_SY]|pA[FVFO_SX]
                mov     ecx, pB

                pfadd(mm0_mm7)
                mov     edx, offs_SX

                movd    mm1, [ecx+edx*4]
                mov     edx, offs_SY

                movq    tmp_pa, mm0
                punpckldq mm1, [ecx+edx*4]             //pB[FVFO_SY]|pB[FVFO_SX]

                mov     ecx, pC
                mov     edx, offs_SX

                pfadd(mm1_mm7)
                movd    mm2, [ecx+edx*4]

                mov     edx, offs_SY
                movq    tmp_pb, mm1

                punpckldq mm2, [ecx+edx*4]             //pA[FVFO_SY]|pA[FVFO_SX]
                pfadd(mm2_mm7)
                movq    tmp_pc, mm2
         }


        //SETFPD( cmdFifo, ghw0->sVx, FLTP(pA)[FVFO_SX] + PIXEL_OFFSET );
        //SETFPD( cmdFifo, ghw0->sVy, FLTP(pA)[FVFO_SY] + PIXEL_OFFSET );
        SETPD( cmdFifo, ghw0->sVx, *((int*)&tmp_pa  ));
        SETPD( cmdFifo, ghw0->sVy, *((int*)&tmp_pa+1));

        if(setupFlag & (SST_SETUP_RGB |
                        SST_SETUP_A))   SETPD( cmdFifo, ghw0->sARGB, aColor );
        if(setupFlag & SST_SETUP_Z)     SETPD( cmdFifo, ghw0->sVz, *((int*)&z1) );
        if(setupFlag & SST_SETUP_Wfbi)  
#ifdef DCT_FIX
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&wb1) );
#else
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&w1) );
#endif
        if(setupFlag & SST_SETUP_W0)    
#ifdef DCT_FIX
           SETPD( cmdFifo, ghw0->sOow0, *((int*)&w1));
#else
           SETPD( cmdFifo, ghw0->sOow0, ((int*)pA)[FVFO_RHW] );
#endif
        if(setupFlag & SST_SETUP_ST0) {
          SETPD( cmdFifo, ghw0->sSow0, *((int*)&s1) );
          SETPD( cmdFifo, ghw0->sTow0, *((int*)&t1) );
        }
      #if (NUMTEXTUREUNITS > 1)
        if(setupFlag & SST_SETUP_ST1) {
          SETPD( cmdFifo, ghw0->sSow1, *((int*)&s1a) );
          SETPD( cmdFifo, ghw0->sTow1, *((int*)&t1a) );
        }
      #endif

        //Send 2nd vertex
        //SETFPD( cmdFifo, ghw0->sVx, FLTP(pB)[FVFO_SX] + PIXEL_OFFSET );
        //SETFPD( cmdFifo, ghw0->sVy, FLTP(pB)[FVFO_SY] + PIXEL_OFFSET );
        SETPD( cmdFifo, ghw0->sVx, *((int*)&tmp_pb  ));
        SETPD( cmdFifo, ghw0->sVy, *((int*)&tmp_pb+1));
        if(setupFlag & (SST_SETUP_RGB |
                        SST_SETUP_A)) SETPD( cmdFifo, ghw0->sARGB, bColor );
        if(setupFlag & SST_SETUP_Z)   SETPD( cmdFifo, ghw0->sVz, *((int*)&z2));
        if(setupFlag & SST_SETUP_Wfbi) 
#ifdef DCT_FIX
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&wb2) );
#else
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&w2) );
#endif
        if(setupFlag & SST_SETUP_W0)  
#ifdef DCT_FIX
           SETPD( cmdFifo, ghw0->sOow0, *((int*)&w2) );
#else
           SETPD( cmdFifo, ghw0->sOow0, ((int*)pB)[FVFO_RHW] );
#endif
        if(setupFlag & SST_SETUP_ST0) {
          SETPD( cmdFifo, ghw0->sSow0, *((int*)&s2) );
          SETPD( cmdFifo, ghw0->sTow0, *((int*)&t2) );
        }
      #if (NUMTEXTUREUNITS > 1)
        if(setupFlag & SST_SETUP_ST1) {
          SETPD( cmdFifo, ghw0->sSow1, *((int*)&s1b) );
          SETPD( cmdFifo, ghw0->sTow1, *((int*)&t1b) );
        }
      #endif

        // Send 3rd vertex
        //SETFPD( cmdFifo, ghw0->sVx, FLTP(pC)[FVFO_SX] + PIXEL_OFFSET );
        //SETFPD( cmdFifo, ghw0->sVy, FLTP(pC)[FVFO_SY] + PIXEL_OFFSET );
        SETPD( cmdFifo, ghw0->sVx, *((int*)&tmp_pc  ));
        SETPD( cmdFifo, ghw0->sVy, *((int*)&tmp_pc+1));
        if(setupFlag & (SST_SETUP_RGB |
                        SST_SETUP_A))   SETPD( cmdFifo, ghw0->sARGB, cColor );
        if(setupFlag & SST_SETUP_Z)     SETPD( cmdFifo, ghw0->sVz, *((int*)&z3) );
        if(setupFlag & SST_SETUP_Wfbi)  
#ifdef DCT_FIX
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&wb3) );
#else
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&w3) );
#endif
        if(setupFlag & SST_SETUP_W0)    
#ifdef DCT_FIX
           SETPD( cmdFifo, ghw0->sOow0, *((int*)&w3) );
#else
           SETPD( cmdFifo, ghw0->sOow0, ((int*)pC)[FVFO_RHW] );
#endif
        if(setupFlag & SST_SETUP_ST0) {
          SETPD( cmdFifo, ghw0->sSow0, *((int*)&s3) );
          SETPD( cmdFifo, ghw0->sTow0, *((int*)&t3) );
        }
      #if (NUMTEXTUREUNITS > 1)
        if(setupFlag & SST_SETUP_ST1) {
          SETPD( cmdFifo, ghw0->sSow1, *((int*)&s1c) );
          SETPD( cmdFifo, ghw0->sTow1, *((int*)&t1c) );
        }
      #endif

      }     //end else      if(ST0_W0_Z_RGB_A)

#if  defined( CMDFIFO) && defined( C_AUTOSTRIP )
      }
      else
      {
      DummyPH = (CMDFIFO_BUILD_PK3( CMD_CONT, 1, setupFlag, 1 ) | ContinueTest);
      SETPH( cmdFifo, DummyPH);
        __asm {
                movd    mm7, pix_offs
                punpckldq mm7, mm7

                mov     ecx, pC
                mov     edx, offs_SX

                pfadd(mm1_mm7)
                movd    mm2, [ecx+edx*4]

                mov     edx, offs_SY

                punpckldq mm2, [ecx+edx*4]             //pc[FVFO_SY]|pA[FVFO_SX]
                pfadd(mm2_mm7)
                movq    tmp_pc, mm2

      }
      // Send 3rd vertex
        SETPD( cmdFifo, ghw0->sVx, *((int*)&tmp_pc  ));
        SETPD( cmdFifo, ghw0->sVy, *((int*)&tmp_pc+1));
        if(setupFlag & (SST_SETUP_RGB |
                        SST_SETUP_A))   SETPD( cmdFifo, ghw0->sARGB, cColor );
        if(setupFlag & SST_SETUP_Z)     SETPD( cmdFifo, ghw0->sVz, *((int*)&z3) );
        if(setupFlag & SST_SETUP_Wfbi)  
#ifdef DCT_FIX
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&wb3) );
#else
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&w3) );
#endif
        if(setupFlag & SST_SETUP_W0)    
#ifdef DCT_FIX
           SETPD( cmdFifo, ghw0->sOow0, *((int*)&w3) );
#else
           SETPD( cmdFifo, ghw0->sOow0, ((int*)pC)[FVFO_RHW] );
#endif
        if(setupFlag & SST_SETUP_ST0) {
          SETPD( cmdFifo, ghw0->sSow0, *((int*)&s3) );
          SETPD( cmdFifo, ghw0->sTow0, *((int*)&t3) );
        }
      #if (NUMTEXTUREUNITS > 1)
        if(setupFlag & SST_SETUP_ST1) {
          SETPD( cmdFifo, ghw0->sSow1, *((int*)&s1c) );
          SETPD( cmdFifo, ghw0->sTow1, *((int*)&t1c) );
        }
      #endif

      }

#endif
      // Specular on Textures
      // specular highlights on texture where the specular color is not black
      if (   (pRc->specular) && (pRc->texture != 0)
           && (CI_MASKALPHA(pA[FVFO_SPECULAR] | pB[FVFO_SPECULAR] | pC[FVFO_SPECULAR]) != 0))
      {
        // if specular is turned on then we need to add in the specular color
        // and this means re-render the triangle gouraud shaded using the specular
        // color information and then add it to the textured triangle we just rendered.
        // We can't just add it to the triangle color because this color would be
        // blended or replaced with the texture and highlights are added into textures.
        //
        ULONG fbzMode, setup;
        ULONG combineModeFBI;
        ULONG fbzColorPath; 

#ifdef C_AUTOSTRIP
        // Because we are doing a second pass for Specular, we need to disable Auto-Stripping
        DoInitialTriangle |= FORCE_SINGLE;  //force individual triangles on next pass
#endif

        //CMDFIFO_CHECKROOM( cmdFifo, ((PH1_SIZE * 7) + 8) + (PH3_SIZE + (4 * 3)) );

        // Where FOG() = fog function
        // FOG(T1 + T2) = AlphaFog * FogColor + (1 - AlphaFog)[T1 +T2]
        //    Pass 1    = AlphaFog * FogColor + (1 - AlphaFog)T1
        //    Pass 2    =                       (1 - AlphaFog)T2
        if (pRc->fogEnable) {
          SETPH( cmdFifo, CMDFIFO_BUILD_PK1( 1, 0, fogMode, 0xF ) );
          SETPD( cmdFifo, ghw->fogMode, (pRc->sst.fogMode & ~SST_FOGMULT) | SST_FOGADD);
        }

        // let's just add in the color ignoring the effect on alpha blending for
        // the time being
        // (Src * 1 + Dst * 1)
        SETPH( cmdFifo, CMDFIFO_BUILD_PK1( 1, 0, alphaMode, 0xF ) );
        SETPD( cmdFifo, ghw->alphaMode, (SST_ENALPHABLEND | (SST_A_ONE << SST_RGBSRCFACT_SHIFT) | (SST_A_ONE << SST_RGBDSTFACT_SHIFT)) );

        if ( pRc->zEnable ) {
          if( pRc->zWriteEnable ) {
            // if z-buffering then this triangle z values equals the values written on pass 1
            fbzMode = (pRc->sst.fbzMode & ~(SST_ZFUNC_LT | SST_ZFUNC_GT)) | SST_ZFUNC_EQ;

            SETPH( cmdFifo, CMDFIFO_BUILD_PK1( 1, 0, fbzMode, 0xF ) );
            SETPD( cmdFifo, ghw->fbzMode, fbzMode );
          }
        }

        setup = setupFlag & ~(SST_SETUP_ST0 | SST_SETUP_W0 | SST_SETUP_ST1 | SST_SETUP_W1);

        // texture mapping off and texture blending off
        if (IS_NAPALM) { //NAPALM_CU
            combineModeFBI = ( SST_CM_ENABLE_TWO_PIXELS_PER_CLOCK | 
                               SST_CM_DISABLE_CHROMA_SUBSTITUTION | 
                               SST_CM_USE_COMBINE_MODE ) & pRc->sst.combineModeFBI; 
            SETPH( cmdFifo,CMDFIFO_BUILD_PK1CHIP( 1, 0, combineMode, 1 ) );
            SETPD( cmdFifo, SST_CHIP(ghw,CHIP_FBI)->combineMode, combineModeFBI );
        } 
        else {
            fbzColorPath = SST_RGBSEL_RGBA;
            if (pRc->subPixel == TRUE) fbzColorPath |= SST_PARMADJUST;

            SETPH( cmdFifo, CMDFIFO_BUILD_PK1( 1, 0, fbzColorPath, 0xF ) );
            SETPD( cmdFifo, ghw->fbzColorPath, fbzColorPath );
        }

        SETPH( cmdFifo, CMDFIFO_BUILD_PK3( 0, 3, (setup | (SST_SETUP_RGB | SST_SETUP_A)), 1 ) );
        __asm{
                mov     eax, pA
                mov     edx, offs_SX

                movd    mm7, pix_offs
                movd    mm0, [eax+edx*4]  // mm0: 0 | FLTP(pA)[FVFO_SX]

                mov     ecx, offs_SY
                punpckldq mm7, mm7        // mm7: PIXEL_OFFSET | PIXEL_OFFSET

                punpckldq mm0, [eax+ecx*4]// mm0: FLTP(pA)[FVFO_SY] | FLTP(pA)[FVFO_SX]
                mov     edx, pB

                pfadd(mm0_mm7)            // mm0: FLTP(pA)[FVFO_SY]+PIXEL_OFFSET | FLTP(pA)[FVFO_SX]+PIXEL_OFFSET
                mov     eax, offs_SX

                mov     ecx, offs_SY
                movd    mm1, [edx+eax*4]  // mm1: 0 | FLTP(pB)[FVFO_SX]

                movq    tmp_pa, mm0
                punpckldq mm1, [edx+ecx*4]// mm1: FLTP(pB)[FVFO_SY] | FLTP(pB)[FVFO_SX]

                mov     edx, pC
                pfadd(mm1_mm7)            // mm1: FLTP(pB)[FVFO_SY]+PIXEL_OFFSET | FLTP(pB)[FVFO_SX]+PIXEL_OFFSET

                mov     eax, offs_SX
                mov     ecx, offs_SY

                movd    mm2, [edx+eax*4]  // mm2: 0 | FLTP(pC)[FVFO_SX]
                movq    tmp_pb, mm1

                punpckldq mm2, [edx+ecx*4]// mm2: FLTP(pC)[FVFO_SY] | FLTP(pC)[FVFO_SX]
                pfadd(mm2_mm7)

                movq    tmp_pc, mm2
        }

        // Send 1st vertex
        //SETFPD( cmdFifo, ghw0->sVx, FLTP(pA)[FVFO_SX] + PIXEL_OFFSET );
        //SETFPD( cmdFifo, ghw0->sVy, FLTP(pA)[FVFO_SY] + PIXEL_OFFSET );
        SETPD( cmdFifo, ghw0->sVx, *((int*)&tmp_pa  ));
        SETPD( cmdFifo, ghw0->sVy, *((int*)&tmp_pa+1));

        SETPD( cmdFifo, ghw0->sARGB, pA[FVFO_SPECULAR] );
        if(setup & SST_SETUP_Z)    SETPD( cmdFifo, ghw0->sVz, *((int*)&z1) );
        if(setup & SST_SETUP_Wfbi) 
#ifdef DCT_FIX
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&wb1) );
#else
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&w1) );
#endif

        // Send 2nd vertex
        //SETFPD( cmdFifo, ghw0->sVx, FLTP(pB)[FVFO_SX] + PIXEL_OFFSET );
        //SETFPD( cmdFifo, ghw0->sVy, FLTP(pB)[FVFO_SY] + PIXEL_OFFSET );
        SETPD( cmdFifo, ghw0->sVx, *((int*)&tmp_pb  ));
        SETPD( cmdFifo, ghw0->sVy, *((int*)&tmp_pb+1));
        SETPD( cmdFifo, ghw0->sARGB, pB[FVFO_SPECULAR] );
        if(setup & SST_SETUP_Z)    SETPD( cmdFifo, ghw0->sVz, *((int*)&z2) );
        if(setup & SST_SETUP_Wfbi) 
#ifdef DCT_FIX
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&wb2) );
#else
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&w2) );
#endif

        // Send 3rd vertex
        //SETFPD( cmdFifo, ghw0->sVx, FLTP(pC)[FVFO_SX] + PIXEL_OFFSET );
        //SETFPD( cmdFifo, ghw0->sVy, FLTP(pC)[FVFO_SY] + PIXEL_OFFSET );
        SETPD( cmdFifo, ghw0->sVx, *((int*)&tmp_pc  ));
        SETPD( cmdFifo, ghw0->sVy, *((int*)&tmp_pc+1));
        SETPD( cmdFifo, ghw0->sARGB, pC[FVFO_SPECULAR] );
        if(setup & SST_SETUP_Z)    SETPD( cmdFifo, ghw0->sVz, *((int*)&z3) );
        if(setup & SST_SETUP_Wfbi) 
#ifdef DCT_FIX
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&wb3) );
#else
           SETPD( cmdFifo, ghw0->sOowfbi, *((int*)&w3) );
#endif

        // restore everything back to the way it was
        if (IS_NAPALM) { //NAPALM_CU
            SETPH( cmdFifo,CMDFIFO_BUILD_PK1CHIP( 1, 0, combineMode, 1 ) );
            SETPD( cmdFifo, SST_CHIP(ghw,CHIP_FBI)->combineMode, pRc->sst.combineModeFBI );
        } 
        else {
            SETPH( cmdFifo, CMDFIFO_BUILD_PK1( 1, 0, fbzColorPath, 0xF ) );
            SETPD( cmdFifo, ghw->fbzColorPath, pRc->sst.fbzColorPath );
        }
        
        if (pRc->fogEnable) {
          SETPH( cmdFifo, CMDFIFO_BUILD_PK1( 1, 0, fogMode, 0xF ) );
          SETPD( cmdFifo, ghw->fogMode, pRc->sst.fogMode );
        }

        SETPH( cmdFifo, CMDFIFO_BUILD_PK1( 2, 1, alphaMode, 0xF ) );
        SETPD( cmdFifo, ghw->alphaMode, pRc->sst.alphaMode );
        SETPD( cmdFifo, ghw->fbzMode, pRc->sst.fbzMode );
      }
    }
#ifdef C_AUTOSTRIP
    else
    {
          // Culled triangle
          // Reset the Auto Strip/Fan Logic
          DoInitialTriangle |= NO_CONTINUE;
          PrevA = PrevB = PrevC = 0;
    }
#endif

  } // every triangle

  __asm{ femms }
  CMDFIFO_EPILOG( cmdFifo );

} // Independat Triangle rendering

#endif // ( DX >= 6 )

#endif // K6_2