/* -*-c++-*- */
/* $Header: E:\Src\Current\3dfx\devel\h5\Win9x\dx\minivdd\dfp.c, 18, 3/8/00 11:02:43 AM, Dan O'Connel$ */
/*
** Copyright (c) 1995-1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished  -
** rights reserved under the Copyright Laws of the United States.
**
** File name:   dfp.c
**
** Description: Support functions for a digital flat panel.
**
** $Revision: 18$
** $Date: 3/8/00 11:02:43 AM$
**
** $Log: 
**  18   3dfx      1.17        03/08/00 Dan O'Connel    Insure that driver doesn't 
**       report to 3dfx Tools that it can support DFP on a Voodoo3.
**  17   3dfx      1.16        02/23/00 Dan O'Connel    Minor clean up of DFP code
**       to prepare for future work.
**  16   3dfx      1.15        02/07/00 Dan O'Connel    Remove a lot of O/S
**       specific ifdefs from these files by changing the way IS_VOODOO3 and
**       IS_NAPALM macros are used.
**  15   3dfx      1.14        02/02/00 Dan O'Connel    Add code to access Bios
**       "Board Config" information to check for TvOut and DFP support.  Also sync
**       up with change to WinNT/Win2K.
**  14   3dfx      1.13        01/21/00 Kyle Pratt      modified panelon paneloff
**       to set and clear SST_VIDEOIN_TV_DATA_SCRAMBLE_DISABLE
**  13   3dfx      1.12        12/30/99 Dan O'Connel    Make source in dfp.c common
**       between Win9x and WinNT/Win2K.
**  12   3dfx      1.11        12/29/99 Dan O'Connel    Change to use macros which
**       provide a common mechanism to access BIOS Scratch Register 2 across the
**       operating systems.
**       Centralize all access to Scratch Register 2 for TvOut in the bt868.c file,
**       so that it will work for all operating systems.
**       Add support for Napalm BIOS in some places it was missing.
**       Some general cleanup and further steps to make code common between O/Ss.
**  11   3dfx      1.10        10/28/99 Dale  Kenaston  	Changed to return the
**       FPFLAG_SCALING in fp_out instead of setting the scaling bit in
**       GLOBALDATA.dwDFPState.
** 
**  10   3dfx      1.9         10/28/99 Dale  Kenaston  	Removed old I2C code.
**       Separated panelDDC_help to form panelDDC_help and panelEDID_help. Also
**       separated panelDDC. Added scaling bit in GLOBALDATA->dwDFPState. Added
**       isPanelScaling function.
** 
**  9    3dfx      1.8         09/29/99 Dale  Kenaston  Added the panelGbl
**       function.
**  8    3dfx      1.7         09/24/99 Dale  Kenaston  Rearranged include files
**       for shared.h. Added support for dwDFPState var in global data in
**       flatPanelPresent. Removed the unused global from flatPanelPhysical. Added
**       support for bios scaling scratch register.
** 
**  7    3dfx      1.6         09/22/99 Ryan Bissell    I2C-related updates
**  6    3dfx      1.5         09/21/99 Dale  Kenaston  Added code for parsing a v2
**       edid's standard modes.
**  5    3dfx      1.4         09/21/99 Dale  Kenaston  Included dfpdisp.h. Moved
**       the ESTABLISHED_TIMING structure to dfpdisp.h and renamed it to DFP_MODE.
**       Removed the global FpFlags; we're now going to use pDev->dfpDisp.FpFlags.
**       Added code to fill the DFP_DISPLAY mode list with the standard modes for
**       edid 1.x. Removed the static fpDims from flatPanelPhysical; we're now
**       going to use the info from pDev->dfpDisp.
**  4    3dfx      1.3         09/16/99 Dale  Kenaston  Renamed NAPALM_HW to
**       NAPALM_HW_DFP
**       Fixed a bug in the isPanelActive where it was reading from d4 instead of
**       d5
**       Added the panelSet function which is a combination of mrtiSet and
**       mrtiConfig, but
**       	without all the mrti code
** 
**  3    3dfx      1.2         09/16/99 Dale  Kenaston  Renamed XLCD_DISPLAY ->
**       DFP_DISPLAY
**       Renamed DfpDisp_* -> DFPDisp_*
**       Renamed xdisp ->disp
** 
**  2    3dfx      1.1         09/14/99 Dale  Kenaston  Added code to set the new
**       napalm dfp resolution scratch register according to the dfp's edid
**  1    3dfx      1.0         09/11/99 StarTeam VTS Administrator 
** $History: dfp.c $
** 
** *****************  Version 3  *****************
** User: Dale         Date: 9/09/99    Time: 6:01p
** Updated in $/devel/h5/Win9x/dx/minivdd
** Included bios.h
** Added the NAPALM_HW define hack to force napalm code use on v3
** Modified panelOn, panelOff, flatPanelPhysical and isPanelActive
**      to use new scratch registers
** Modified flatPanelPhysical to not set the bios scratch register
** Removed #if 0 around mrtiPresent call in flatPanelPresent
** 
** *****************  Version 2  *****************
** User: Rbissell     Date: 9/07/99    Time: 2:15p
** Updated in $/devel/h5/Win9x/dx/minivdd
** Deployed the new modularized I2C code.
** 
** *****************  Version 1  *****************
** User: Dalev        Date: 8/20/99    Time: 4:26p
** Created in $/devel/h5/Win9x/dx/minivdd
** Original;  This file contains all code necessary to communicate with a
** DFP..
** 
**
** 
*/

#if 1

#ifndef WINNT
   #define FAR
   #define WINAPI __stdcall
   typedef          char  CHAR;
   typedef   signed char  SCHAR;
   typedef unsigned char  UCHAR;

   typedef          short SHORT;
   typedef unsigned short USHORT;

   typedef          long  LONG;
   typedef unsigned long  ULONG;

   typedef unsigned char  BYTE,  FAR* LPBYTE;
   typedef unsigned short WORD,  FAR* LPWORD;
   typedef unsigned long  DWORD, FAR* LPDWORD;
   typedef          void  VOID,  FAR* LPVOID;
   typedef          int   BOOL;
   typedef int (FAR WINAPI *FARPROC)();


   #pragma pack( 1 )
       #include "gdidefs.h"
       #include "dibeng.h"
   #pragma pack()
#endif

   #include "3dfx.h"

#ifdef WINNT
   #include "miniport.h"
   #include "ntddvdeo.h"
   #include "video.h"
   #include "h3.h"
   #include "dfp.h"
   #define DELAY(usecs)      VideoPortStallExecution(usecs)

#else
   #include "h3vdd.h"
   #include "h3.h"

   #include "shared.h"
   #define VDDONLY
   #include "devtable.h"
   #include "h3g.h"
   #undef  VDDONLY

      #include "i2c/di_i2c.h"
      #define DELAY(usecs) CM_Yield(usecs, CM_YIELD_RESUME_EXEC);
#endif  // else #ifdef WINNT

#else    // debug on PC

   #define PDEVTABLE void*
   #define XLCD_DEBUG   1

#endif

#ifdef WINNT
#define DFP_DEVINFOPTR(pcontext) ((PHW_DEVICE_EXTENSION)(pcontext))
#define IS_NAPALM_X(pcontext)   (0x06 <= (DFP_DEVINFOPTR(pcontext))->PCIDeviceID)
#define IS_VOODOO3_X(pcontext)  (0x05 >= (DFP_DEVINFOPTR(pcontext))->PCIDeviceID)
#else
#define DFP_DEVINFOPTR(pcontext) ((PDEVTABLE)(pcontext))
#define IS_NAPALM_X(pcontext)   (IS_NAPALM((DFP_DEVINFOPTR(pContext))->dwVendorDeviceID))
#define IS_VOODOO3_X(pcontext)  (IS_VOODOO3((DFP_DEVINFOPTR(pContext))->dwVendorDeviceID))
#endif

#include "dfp.h"
#include "dfpdisp.h"
#ifndef WINNT  //xlcd Not Ported From Win9x
#include "xlcd.h"
#endif

#include "bios.h"

// established timings in EDID per VESA

static DFP_MODE EstTimings[] =
   {{800, 600, 60},
	{800, 600, 56},
	{640, 480, 75},
	{640, 480, 72},
	{640, 480, 67},
	{640, 480, 60},
	{720, 400, 88},
	{720, 400, 70},
	{1280, 1024, 75},
	{1024, 768, 75},
	{1024, 768, 70},
	{1024, 768, 60},
	{1024, 768, 87},
	{832, 624, 75},
	{800, 600, 75},
	{800, 600, 72},
	{1152, 700, 75},
	{0, 0, 0}};


const DFP_DISPLAY DFPdisp_1024 = {1344, 1024, 160, 24, 806, 768, 6, 3};
const DFP_DISPLAY DFPdisp_1280 = {1680, 1280, 112, 48, 1066, 1024, 6, 3};

#undef PNL_1280_TEST   // to test 12x10 panel with no EDID, best we can do



/*----------------------------------------------------------------------
Macro name:  READSCRATCHREGISTER3

Description:    read a byte value from BIOS scratch register 3 (location 0x1F)

Information:    
----------------------------------------------------------------------*/
#ifdef WINNT
#define READSCRATCHREGISTER3(pContext,result)  \
    { \
    /* save index reg setting */ \
    FxU8 bSavedIndex = VideoPortReadPortUchar((PUCHAR)DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_INDEX] + 0xd4); \
    /* return value from register shared with BIOS */ \
    VideoPortWritePortUchar((PUCHAR)DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_INDEX] + 0xd4, 0x1f);  /* select the reg */ \
    result = VideoPortReadPortUchar((PUCHAR)DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_INDEX] + 0xd5);  /* read the reg */ \
    /* restore index reg setting */ \
    VideoPortWritePortUchar((PUCHAR)DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_INDEX] + 0xd4, bSavedIndex); \
    }
#else
#define READSCRATCHREGISTER3(pContext,result)  \
    { \
    /* save index reg setting*/ \
    FxU8 bSavedIndex = inp ((FxU16)(DFP_DEVINFOPTR(pContext)->IoBase + 0xd4)); \
    /* return value from register shared with BIOS */ \
    outp ((FxU16)(DFP_DEVINFOPTR(pContext)->IoBase + 0xd4), 0x1f);  /* select the reg */ \
    result = inp ((FxU16)(DFP_DEVINFOPTR(pContext)->IoBase + 0xd5));  /* read the reg */ \
    /* restore index reg setting */ \
    outp ((FxU16)(DFP_DEVINFOPTR(pContext)->IoBase + 0xd4), bSavedIndex); \
    }
#endif
/*----------------------------------------------------------------------
Macro name:  WRITESCRATCHREGISTER3

Description:    write a byte value to BIOS scratch register 3  (location 0x1F)

Information:    
----------------------------------------------------------------------*/
#ifdef WINNT
#define WRITESCRATCHREGISTER3(pContext,value) \
{ \
    /* save index reg setting */ \
    FxU8 bSavedIndex = VideoPortReadPortUchar((PUCHAR)DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_INDEX] + 0xd4); \
    /* write value to register shared with BIOS */ \
    VideoPortWritePortUchar((PUCHAR)DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_INDEX] + 0xd4, 0x1f);  /* select the reg */ \
    VideoPortWritePortUchar((PUCHAR)DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_INDEX] + 0xd5,value);  /* write the reg */ \
    /* restore index reg setting */ \
    VideoPortWritePortUchar((PUCHAR)DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_INDEX] + 0xd4, bSavedIndex); \
}
#else
#define WRITESCRATCHREGISTER3(pContext,value) \
{ \
    /* save index reg setting */ \
    FxU8 bSavedIndex = inp ((FxU16)(DFP_DEVINFOPTR(pContext)->IoBase + 0xd4)); \
    /* write value to register shared with BIOS */ \
    outp ((FxU16)(DFP_DEVINFOPTR(pContext)->IoBase + 0xd4), 0x1f);  /* select the reg */ \
    outp ((FxU16)(DFP_DEVINFOPTR(pContext)->IoBase + 0xd5), value);  /* write the reg */ \
    /* restore index register setting */ \
    outp ((FxU16)(DFP_DEVINFOPTR(pContext)->IoBase + 0xd4), bSavedIndex); \
}
#endif


/*----------------------------------------------------------------------
Function name:  panelEDID_help and panelEDID

Description:    Retrieves the panel EDID Info.

Information:    

Return:         INT     0 or -1.
----------------------------------------------------------------------*/


int panelEDID_help(void * pContext, FxU8 bAddress, FxU8 *ddcData)
{
#ifdef WINNT
    H3_MEMBASE0 *sstIO = (PH3_MEMBASE0) DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_REGS_INDEX];
#else
	SstIORegs *sstIO = (void *)DFP_DEVINFOPTR(pContext)->RegBase[HWINFO_SST_IOREGS_INDEX];
#endif
   I2CKEY dfpddc;
	int i, nb, sum, retstat;
	DFP_MODE *eTiming = 0;
	FxU8 retries = 2;

retry:
   
   dfpddc = i2c_getaccess(pContext, I2C_FLATPANELDDC, I2C_NORMALSPEED);
   if (dfpddc == I2C_NOTAKEY)
      goto bailout;

   retstat = 0;
   retstat |= !i2c_start(pContext, dfpddc);
   retstat |= !i2c_sendbyte(pContext, dfpddc, (FxU8)(bAddress | I2C_WRITESLAVE));
   retstat |= !i2c_sendbyte(pContext, dfpddc, 0);
   retstat |= !i2c_stop(pContext, dfpddc);
   retstat |= !i2c_start(pContext, dfpddc);
   retstat |= !i2c_sendbyte(pContext, dfpddc, (FxU8)(bAddress | I2C_READSLAVE));

   if (retstat)
      goto bailout; // it's obvious this retry won't work; try to speed things up


#define LCD_BUG_PRINCETONDFP
#ifdef LCD_BUG_PRINCETONDFP
   // @RBISSELL: The Princeton "directdigital" DFP returns a bogus EDID header.
   //            Instead of "00,FF,FF,FF,FF,FF,FF,00", it starts the EDID with the
   //            the header "00,00,FF,FF,FF,FF,FF,00".  The checksum reported within
   //            the EDID is consistent with this error.

   *(FxU32*)&ddcData[0] = 0x00000000;
   for (sum = i = 0; i < 4; i++)
	{
      i2c_readbyte(pContext, dfpddc, &ddcData[i], 1);
		sum += ddcData[i];
		DELAY(10000);
	}
   if ((*(FxU32*)&ddcData[0]) == 0xFFFF0000)  // is it a broken Princeton DFP?
   {
      // change the second byte to 0xFF, and adjust the checksum accordingly.
      sum += 0xFF;
      *(FxU32*)&ddcData[0] = 0xFFFFFF00;
   }
   for (nb = 128; i < nb; i++) //read the remainder of the edid
#else
	for (sum = i = 0, nb = 128; i < nb; i++)
#endif
   {
      i2c_readbyte(pContext, dfpddc, &ddcData[i], i != (nb - 1));
		nb = ddcData[0] >= 0x20 ? 256 : 128;
		sum += ddcData[i];
		DELAY(10000);
	}

bailout:
   i2c_stop(pContext, dfpddc);
   i2c_endaccess(pContext, dfpddc);

	if (!retstat && !(sum & 255) && ddcData[0] < 0x80 && sum)
      return 0;
	else if (--retries)
		goto retry;
	else
      return -1;
}

int panelEDID(void * pContext, FxU8 *ddcData)
{
   if (panelEDID_help(pContext, 0xA2, ddcData))
      if (panelEDID_help(pContext, 0xA0, ddcData))
         return -1;

   return 0;
}



/*----------------------------------------------------------------------
Function name:  panelDDC_help and panelDDC

Description:    Parse the panel EDID Info.

Information:    

Return:         INT     0 or -1.
----------------------------------------------------------------------*/


int panelDDC_help(void * pContext, DFP_DISPLAY *disp, FxU8 bAddress)
{
#ifdef WINNT
    H3_MEMBASE0 *sstIO = (PH3_MEMBASE0) DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_REGS_INDEX];
#else
    SstIORegs *sstIO = (void *)DFP_DEVINFOPTR(pContext)->RegBase[HWINFO_SST_IOREGS_INDEX];
#endif
	int i, j, blk, clk, retstat;
    FxU8 ddcData[256], *dt;
    DFP_MODE *eTiming = 0;
    FxU8 retries = 2;

retry:

   if (panelEDID_help(pContext, bAddress, ddcData) != -1)
	{
		/* EDID parsing is taken from VESA EDID Standard v3 r0, 1997 */

		if (ddcData[0] >= 0x20)   // decode V2.X EDID
		{
			dt = &ddcData[128];   
			// skip over luma table if present
			if (ddcData[0x7e] & 0x20)
			{
				if (ddcData[0x80] & 0x80)
					dt += ((ddcData[0x80] & 0x1f) * 3);
				else
					dt += (ddcData[0x80] & 0x1f);
			}
			// skip frequency ranges
			dt += (8 * ((ddcData[0x7e] >> 2) & 7));
			// if timing ranges are present
			if (ddcData[0x7e] & 3)
			{
				// extract H parameters
				blk = ((dt[4] << 4) & 0xf00) | dt[2];
				blk += (((dt[13] << 4) & 0xf00) | dt[11]);
				disp->dispWidth = ((dt[23] << 4) & 0xf00) | dt[21];
				disp->htotal = disp->dispWidth + (blk / 2);
				disp->hsyncOfst = ((dt[8] << 2) & 0x300) | dt[5];
				disp->hsyncWidth = ((dt[8] << 4) & 0x300) | dt[6];
				// extract V parameters
				blk = ((dt[4] << 8) & 0xf00) | dt[3];
				blk += ((dt[13] << 8) & 0xf00) | dt[12];
				disp->dispHeight = ((dt[23] << 8) & 0xf00) | dt[22];
				disp->vtotal = disp->dispHeight + (blk / 2);
				disp->vsyncOfst = ((dt[8] << 2) & 0x030) | (dt[7] >> 4);
				disp->vsyncWidth = ((dt[8] << 4) & 0x030) | (dt[7] & 15);
				// get the average clk rate
				clk = *(FxU16 *)&dt[0];
				clk += (dt[9] + (dt[10] << 8));
				disp->dispClock = (long)(clk / 2) * (long)10;
			}
         // skip detailed range limits
         dt += 27 * (ddcData[0x7e] & 3);
         // if the 4-byte timing codes are present
         if (ddcData[0x7f] & 0xf8)
         {
            for(i=0; i<(ddcData[0x7f] & 0xf8); i++)
            {
               DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[i].width = (dt[0] * 16) + 256;
               switch(dt[2])
               {
                  case 160: // 16:10 Aspect Ratio
                     DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[i].height = (DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[i].width * 10) / 16;
                     break;
                  case 133: //  4: 3 Aspect Ratio
                     DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[i].height = (DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[i].width * 3 ) /  4;
                     break;
                  case 125: //  5: 4 Aspect Ratio
                     DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[i].height = (DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[i].width * 4 ) /  5;
                     break;
                  case 178: // 16: 9 Aspect Ratio
                     DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[i].height = (DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[i].width * 9 ) / 16;
                     break;
               }
               DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[i].refresh = dt[3];
               if (i == MAX_DFP_MODES)
                  break;

               dt += 4;
            }
         }
		}
		else    // EDID V1.X
		{
			// parse out the established timings to get the minimum we need to
			// know - display height, width and refresh.  Give preference to 
			// highest resolution and allow only 60hz refresh.

         j=0;
			for (i = 0; i < 8; i++)
			{
				if ((ddcData[0x24] & (1 << i)) &&
					EstTimings[i + 8].refresh == 60)
            {
						DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j++] = EstTimings[i + 8];
                  if (j == MAX_DFP_MODES)
                     goto mode_list_full;
            }
			}
			for (i = 0; i < 8; i++)
			{
				if ((ddcData[0x23] & (1 << i)) &&
					EstTimings[i + 0].refresh == 60)
            {
						DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j++] = EstTimings[i + 0];
                  if (j == MAX_DFP_MODES)
                     goto mode_list_full;
            }
			}
         for (i = 0; i < 16; i+=2)
         {
            if((ddcData[0x26 + i]     != 0x01) || //STI is valid
               (ddcData[0x26 + i + 1] != 0x01))
            {
               if((ddcData[0x26 + i + 1] & 0x3f) == 0) //Refresh rate == 60Hz
               {
                  DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j].width = (ddcData[0x26 + i] + 31) * 8;
                  switch(ddcData[0x26 + i + 1] & 0xc0)
                  {
                  case 0x00: // 16:10 Aspect Ratio
                     DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j].height = (DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j].width * 10) / 16;
                     break;
                  case 0x40: //  4: 3 Aspect Ratio
                     DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j].height = (DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j].width * 3 ) /  4;
                     break;
                  case 0x80: //  5: 4 Aspect Ratio
                     DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j].height = (DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j].width * 4 ) /  5;
                     break;
                  case 0xc0: // 16: 9 Aspect Ratio
                     DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j].height = (DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j].width * 9 ) / 16;
                     break;
                  }
                  DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[j].refresh = 60;
                  j++;
                  if (j == MAX_DFP_MODES)
                     goto mode_list_full;
               }
            }
         }
mode_list_full:

			dt = &ddcData[54];
			if ((clk = *(FxU16 *)&dt[0]) <= 0x0101)
			{
				disp->vsyncWidth = 6;    // HACK HACK
				return (0);    // no detailed timing, just dimensions
			}
			disp->dispClock = clk * 10;
			// extract H parameters
			disp->dispWidth = ((dt[4] << 4) & 0xf00) | dt[2];
			disp->htotal = ((dt[4] << 8) & 0xf00) + dt[3] + disp->dispWidth;
			disp->hsyncOfst = ((dt[11] << 2) & 0x300) | dt[8];
			disp->hsyncWidth = ((dt[11] << 4) & 0x300) | dt[9];
			// extract V parameters
			disp->dispHeight = ((dt[7] << 4) & 0xf00) | dt[5];
			disp->vtotal = ((dt[7] << 8) & 0xf00) + dt[6] + disp->dispHeight;
			disp->vsyncOfst = ((dt[11] << 2) & 0x30) | (dt[10] >> 4);
			disp->vsyncWidth = ((dt[11] << 4) & 0x30) | (dt[10] & 15);
		}

        retstat = 0;
	}
	else if (--retries)
		goto retry;
	else
		retstat = 1;

	return (retstat ? -1 : 0);
}

/*----------------------------------------------------------------------
Function name:  panelDDC

Description:    Call 'panelDDC_help()' to get panel EDID.

Information:    

Return:         VOID
----------------------------------------------------------------------*/
int panelDDC(void * pContext, DFP_DISPLAY *disp)
{
    if (panelDDC_help(pContext, disp, 0xA2))
        if (panelDDC_help(pContext, disp, 0xA0))
            return -1;

        return 0;
}

/*----------------------------------------------------------------------
Function name:  panelOff

Description:    Turn off the panel device.

Information:    

Return:         VOID
----------------------------------------------------------------------*/
void panelOff(void * pContext)
{
    FxU8 dfpbits;
#ifdef WINNT
    H3_MEMBASE0 *sstIOregs = (PH3_MEMBASE0) DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_REGS_INDEX];
#else
    SstIORegs *sstIOregs = (void *)DFP_DEVINFOPTR(pContext)->RegBase[HWINFO_SST_IOREGS_INDEX];
#endif
    sstIOregs->vidInFormat &= ~(SST_VIDEOIN_TVOUT_ENABLE);  // off
    sstIOregs->vidInFormat |= SST_VIDEOIN_G4_FOR_POSEDGE;  // Like Brooktree.

    // If its not V3 then clear the bit so we don't break TVout code.
    if(!IS_VOODOO3_X(pContext)) 
    {
       sstIOregs->vidInFormat &= ~SST_VIDEOIN_TV_DATA_SCRAMBLE_DISABLE;
    }

    if(NAPALM_HW_DFP && IS_VOODOO3_X(pContext))
    {
        READSCRATCHREGISTER2(pContext, dfpbits);
        dfpbits &= ~4;
        WRITESCRATCHREGISTER2(pContext, dfpbits);
    }
    else
    {
        READSCRATCHREGISTER3(pContext, dfpbits);
        dfpbits &= ~BIOS_CRx1F_DFPACTIVE;
        WRITESCRATCHREGISTER3(pContext, dfpbits);
    }

#ifdef WINNT
    // set the timing again because the DFP timing may be different from the monitor timing
    H3SetModeResetTiming(DFP_DEVINFOPTR(pContext));
#endif
}


/*----------------------------------------------------------------------
Function name:  panelOn

Description:    Turn on the panel.

Information:    

Return:         VOID
----------------------------------------------------------------------*/
void panelOn(void * pContext)
{
    FxU8 dfpbits;
#ifdef WINNT
    H3_MEMBASE0 *sstIOregs = (PH3_MEMBASE0) DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_REGS_INDEX];
#else
    SstIORegs *sstIOregs = (void *)DFP_DEVINFOPTR(pContext)->RegBase[HWINFO_SST_IOREGS_INDEX];
#endif
    sstIOregs->vidInFormat |= SST_VIDEOIN_TVOUT_ENABLE;
    sstIOregs->vidInFormat &= ~SST_VIDEOIN_G4_FOR_POSEDGE;  // Like Chrontel.

    // If its not V3 then set the bit for direct connect to SII part.
    if(!IS_VOODOO3_X(pContext)) 
    {
       sstIOregs->vidInFormat |= SST_VIDEOIN_TV_DATA_SCRAMBLE_DISABLE;
    }

    if(NAPALM_HW_DFP && IS_VOODOO3_X(pContext))
    {
        READSCRATCHREGISTER2(pContext, dfpbits);
        dfpbits |= 4;
        WRITESCRATCHREGISTER2(pContext, dfpbits);
    }
    else
    {
        READSCRATCHREGISTER3(pContext, dfpbits);
        dfpbits |= BIOS_CRx1F_DFPACTIVE;
        WRITESCRATCHREGISTER3(pContext, dfpbits);
    }

#ifdef WINNT
    // set the timing again because the DFP timing may be different from the monitor timing
    H3SetModeResetTiming(DFP_DEVINFOPTR(pContext));
#endif WINNT

}

/*----------------------------------------------------------------------
Function name:  flatPanelPresent

Description:    Query for a panel device.

Information:    

Return:         INT     0 for no panel device, else panel device
----------------------------------------------------------------------*/
int flatPanelPresent(void * pContext)
{
    I2CKEY dfpddc;

    int status = 1;

#ifndef DANOTESTDFP
// No released Voodoo3 boards are supported by this driver, but there are a few prototype Voodoo3 DFP
// are used in testing.  These prototype boards can be enabled by defining DANOTESTDFP.
    if (IS_VOODOO3_X(pContext))
        return(0);
#endif
    // test if BIOS supports DFP
    if ((IS_NAPALM_X(pContext)) && ((DFP_DEVINFOPTR(pContext)->biosBoardConfigInfo & BIOS_BOARDCONFIG_DFPSUPPORT)==0))
        return(0);

    dfpddc = i2c_getaccess(pContext, I2C_FLATPANELDDC, I2C_NORMALSPEED);
    if (dfpddc == I2C_NOTAKEY)
        return 0; // if we couldn't get a handle for the DFP DDC bus, assume this isn't a DFP board.

    status &= i2c_start(pContext, dfpddc);
    status &= i2c_sendbyte(pContext, dfpddc, (FxU8)(0xa0 | I2C_WRITESLAVE));
    status &= i2c_stop(pContext, dfpddc);
    i2c_endaccess(pContext, dfpddc);


#ifndef WINNT  //xlcd Not Ported From Win9x
    // If panel is connected & we're on a VOODOO3, then make sure
    // a mrti controller chip is on the board.
    if (status && NAPALM_HW_DFP && IS_VOODOO3_X(pContext))
    {
        status &= mrtiPresent(pContext);
    }

    if(DFP_DEVINFOPTR(pContext)->lpDriverData)
        ((GLOBALDATA *)DFP_DEVINFOPTR(pContext)->lpDriverData)->dwDFPState |= status ? 0x1 : 0x0;
#endif

    return (status);
}


#ifdef WINNT
/*----------------------------------------------------------------------
Function name:  flatPanelPhysical

Description:    

Information:    

Return:         INT     0 or -1.
----------------------------------------------------------------------*/
int flatPanelPhysical(void * pContext, FlatPnlPhysical *fpp)
{
#ifndef PNL_1280_TEST
    DFP_DISPLAY disp = DFPdisp_1024;
#else
    DFP_DISPLAY disp = DFPdisp_1280;
#endif
    FxU8 ucTemp;

    // Put the following variables in a #define.
    // Note:  I guess this variable is to be used for mode changes.  During
    // a system boot, fpp->fpFlags will be 0, but on subsequent mode changes
    // FPFLAG_GET_DIMS & FPFLAG_PRESENT should be set and thus eliminate the
    // need to aquire the DFP edid.  dav.
    FxU32 ddcFlags = FPFLAG_GET_DIMS | FPFLAG_PRESENT;

#ifdef newfromWin9x
	disp.dispWidth = 0;
#endif
    fpp->fpFlags &= ~(FPFLAG_PRESENT_AT_BOOT | FPFLAG_PRESENT);

    if (flatPanelPresent (pContext))
    {
        fpp->fpFlags |= FPFLAG_PRESENT;

#if 0
            if(NAPALM_HW_DFP && IS_VOODOO3_X(pContext))
            {
                READSCRATCHREGISTER2(pContext, ucTemp);
                if (ucTemp & 4)
                    fpp->fpFlags |= FPFLAG_PRESENT_AT_BOOT;
            }
            else
            {
                READSCRATCHREGISTER3(pContext, ucTemp);
                if (ucTemp & BIOS_CRx1F_DFPACTIVE)
                    fpp->fpFlags |= FPFLAG_PRESENT_AT_BOOT;
            }
#else
            if ((DFP_DEVINFOPTR(pContext))->bDfpActiveAtBoot)
                fpp->fpFlags |= FPFLAG_PRESENT_AT_BOOT;
#endif //if 0
    }

    if ((fpp->fpFlags & ddcFlags) == ddcFlags)
    {
        if (DFP_DEVINFOPTR(pContext)->dfpDisp.dispWidth)    // if dfpDisp != 0, then we must have EDID.
        {
            fpp->maxWidth = DFP_DEVINFOPTR(pContext)->dfpDisp.dispWidth;
            fpp->maxHeight = DFP_DEVINFOPTR(pContext)->dfpDisp.dispHeight;
            fpp->refreshRate = 60;
            if (isPanelScaling(pContext))
                fpp->fpFlags |= FPFLAG_SCALING;
            return (0);
        }  // Now lets get and return the dimensions of our Panel.
        else if (!panelDDC (pContext, &DFP_DEVINFOPTR(pContext)->dfpDisp))
        {
            if (!DFP_DEVINFOPTR(pContext)->dfpDisp.dispWidth)
                return (-1);

            //Fill in the return values
            fpp->maxWidth = DFP_DEVINFOPTR(pContext)->dfpDisp.dispWidth;
            fpp->maxHeight = DFP_DEVINFOPTR(pContext)->dfpDisp.dispHeight;
            fpp->refreshRate = 60;
            if (isPanelScaling(pContext))
                fpp->fpFlags |= FPFLAG_SCALING;

            //Set the napalm scratch register
            if(!NAPALM_HW_DFP || IS_NAPALM_X(pContext))
            {
                FxU8 dfpbits;

                READSCRATCHREGISTER3(pContext, dfpbits);

                //Set up the res bit field
                if(fpp->maxWidth >= 1600)
                {
                    dfpbits &= ~BIOS_CRx1F_DFPCONNECTED;
                    dfpbits |= BIOS_CRx1F_1600x1200;
                }
                else if(fpp->maxWidth >= 1280)
                {
                    dfpbits &= ~BIOS_CRx1F_DFPCONNECTED;
                    dfpbits |= BIOS_CRx1F_1280x1024;
                }
                else
                {
                    dfpbits &= ~BIOS_CRx1F_DFPCONNECTED;
                    dfpbits |= BIOS_CRx1F_1024x768;
                }

                //Set up the scaling bit
                if (isPanelScaling(pContext))
                    dfpbits |= BIOS_CRx1F_PANELSCALING;
                else
                    dfpbits &= ~BIOS_CRx1F_PANELSCALING;

                WRITESCRATCHREGISTER3(pContext, dfpbits);
                }

                return (0);
        }
        else   // If we can't get EDID; assume 10x7 panel for now.
        {
#ifndef PNL_1280_TEST
            fpp->maxWidth = 1024;
            fpp->maxHeight = 768;
#else
            fpp->maxWidth = 1280;
            fpp->maxHeight = 1024;
#endif
            fpp->refreshRate = 60;

            if(!NAPALM_HW_DFP || IS_NAPALM_X(pContext))
            {
                FxU8 dfpbits;
                READSCRATCHREGISTER3(pContext, dfpbits);
                dfpbits &= ~BIOS_CRx1F_DFPCONNECTED;
                dfpbits |= BIOS_CRx1F_1024x768;
                WRITESCRATCHREGISTER3(pContext, dfpbits);
            }

#ifdef WINNT  // why is this needed DanO 11/1/99???
            if (isPanelActive(pContext))
                fpp->fpFlags |= FPFLAG_ENABLED;
#endif

            return (0);
        }
    }

    return (-1);
}
#else
// the Win9x version follows:
/*----------------------------------------------------------------------
Function name:  flatPanelPhysical

Description:    

Information:    

Return:         INT     0 or -1.
----------------------------------------------------------------------*/
int flatPanelPhysical(void * pContext, FlatPnlPhysical *fpp)
{
    FxU8 ucTemp;

#ifndef PNL_1280_TEST
	DFP_DISPLAY disp = DFPdisp_1024;
#else
	DFP_DISPLAY disp = DFPdisp_1280;
#endif
// Put the following variables in a #define.
// Note:  I guess this variable is to be used for mode changes.  During
// a system boot, fpp->fpFlags will be 0, but on subsequent mode changes
// FPFLAG_GET_DIMS & FPFLAG_PRESENT should be set and thus eliminate the
// need to aquire the DFP edid.  dav.
	DWORD ddcFlags = FPFLAG_GET_DIMS | FPFLAG_PRESENT;

	disp.dispWidth = 0;
	fpp->fpFlags &= ~(FPFLAG_PRESENT_AT_BOOT | FPFLAG_PRESENT);

	if (flatPanelPresent (pContext))
	{
	   fpp->fpFlags |= FPFLAG_PRESENT;

      if(NAPALM_HW_DFP && IS_VOODOO3_X((pContext)))
      {
           READSCRATCHREGISTER2(pContext, ucTemp);
           if (ucTemp & 4)
               fpp->fpFlags |= FPFLAG_PRESENT_AT_BOOT;
      }
      else
      {
           READSCRATCHREGISTER3(pContext, ucTemp);
           if (ucTemp & BIOS_CRx1F_DFPACTIVE)
               fpp->fpFlags |= FPFLAG_PRESENT_AT_BOOT;
      }
	}

	if ((fpp->fpFlags & ddcFlags) == ddcFlags)
	{
		if (DFP_DEVINFOPTR(pContext)->dfpDisp.dispWidth)    // if dfpDisp != 0, then we must have EDID.
		{
			fpp->maxWidth = DFP_DEVINFOPTR(pContext)->dfpDisp.dispWidth;
			fpp->maxHeight = DFP_DEVINFOPTR(pContext)->dfpDisp.dispHeight;
			fpp->refreshRate = 60;
            if (isPanelScaling(pContext))
                fpp->fpFlags |= FPFLAG_SCALING;
			return (0);
		}  // Now lets get and return the dimensions of our Panel.
		else if (!panelDDC (pContext, &DFP_DEVINFOPTR(pContext)->dfpDisp))
		{
			if (!DFP_DEVINFOPTR(pContext)->dfpDisp.dispWidth)
				return (-1);

         //Fill in the return values
			fpp->maxWidth = DFP_DEVINFOPTR(pContext)->dfpDisp.dispWidth;
			fpp->maxHeight = DFP_DEVINFOPTR(pContext)->dfpDisp.dispHeight;
			fpp->refreshRate = 60;
            if (isPanelScaling(pContext))
                fpp->fpFlags |= FPFLAG_SCALING;

            //Set the napalm scratch register
            if(!NAPALM_HW_DFP || IS_NAPALM_X(pContext))
            {
               FxU8 dfpbits;

               READSCRATCHREGISTER3(pContext, dfpbits);

               //Set up the res bit field
               if(fpp->maxWidth >= 1600)
               {
                  dfpbits &= ~BIOS_CRx1F_DFPCONNECTED;
                  dfpbits |= BIOS_CRx1F_1600x1200;
               }
               else if(fpp->maxWidth >= 1280)
               {
                  dfpbits &= ~BIOS_CRx1F_DFPCONNECTED;
                  dfpbits |= BIOS_CRx1F_1280x1024;
               }
               else
               {
                  dfpbits &= ~BIOS_CRx1F_DFPCONNECTED;
                  dfpbits |= BIOS_CRx1F_1024x768;
               }

               //Set up the scaling bit
               if (isPanelScaling(pContext))
                  dfpbits |= BIOS_CRx1F_PANELSCALING;
               else
                  dfpbits &= ~BIOS_CRx1F_PANELSCALING;

               WRITESCRATCHREGISTER3(pContext, dfpbits);
            }

			return (0);
		}
		else   // If we can't get EDID; assume 10x7 panel for now.
		{
#ifndef PNL_1280_TEST
			fpp->maxWidth = 1024;
			fpp->maxHeight = 768;
#else
			fpp->maxWidth = 1280;
			fpp->maxHeight = 1024;
#endif
			fpp->refreshRate = 60;

            if(!NAPALM_HW_DFP || IS_NAPALM_X(pContext))
            {
               FxU8 dfpbits;

               READSCRATCHREGISTER3(pContext, dfpbits);
               dfpbits &= ~BIOS_CRx1F_DFPCONNECTED;
               dfpbits |= BIOS_CRx1F_1024x768;
               WRITESCRATCHREGISTER3(pContext, dfpbits);
            }

			return (0);
		}
	}

	return (-1);
}
#endif
/*----------------------------------------------------------------------
Function name:  isPanelActive

Description:    Retrieve 'PANEL IS Active' flag from 
                BIOS scratch pad register.

Information:    Checks to see if BIOS believe the panel is active.

Return:         FxBool     TRUE or FALSE.
----------------------------------------------------------------------*/
FxBool isPanelActive(void * pContext)
{
    FxU8 ucTemp;

#ifndef DANOTESTDFP
// No released Voodoo3 boards are supported by this driver, but there are a few prototype Voodoo3 DFP
// are used in testing.  These prototype boards can be enabled by defining DANOTESTDFP.
    if (IS_VOODOO3_X(pContext))
        return(FALSE);
#endif

    if(NAPALM_HW_DFP && IS_VOODOO3_X(pContext))
    {
        READSCRATCHREGISTER2(pContext, ucTemp);
        return (!!(ucTemp & 4));      
    }
    else
    {
        READSCRATCHREGISTER3(pContext, ucTemp);
        return (!!(ucTemp & BIOS_CRx1F_DFPACTIVE));
    }
}

/*----------------------------------------------------------------------
Function name:  isPanelScaling

Description:    Look in the parsed EDID data to see if the panel
                supports more than one mode

Information:    Checks to see if BIOS has a scaling panel connected.

Return:         FxBool  TRUE or FALSE
----------------------------------------------------------------------*/
FxBool isPanelScaling(void * pContext)
{
    return DFP_DEVINFOPTR(pContext)->dfpDisp.stdModes[1].width != 0;
}

#ifndef	WINNT
/*----------------------------------------------------------------------
Function name:  panelFixupVGA

Description:    Called when disabling desktop to fix VGA CRTC registers FBO
                a proper full screen DOS box on the panel.

Information:    Only called if panel is active.

Return:         Returns 0
----------------------------------------------------------------------*/

int panelFixupVGA(void * pContext, FxU32 unused)
{
    outpw ((FxU16)(DFP_DEVINFOPTR(pContext)->IoBase + 0xd4), 0x0c11);   // unlock VGA registers
    outpw ((FxU16)(DFP_DEVINFOPTR(pContext)->IoBase + 0xd4), 0xff06);   // adjust refresh rate
    outpw ((FxU16)(DFP_DEVINFOPTR(pContext)->IoBase + 0xd4), 0x8c11);   // relock VGA registers
    return (0);
}
#endif

/*----------------------------------------------------------------------
Function name:  panelSet

Description:    A stripped down version of mrtiSet for non-mrti devices

Information:    

Return:         INT     0 or -1.
----------------------------------------------------------------------*/
int panelSet(void * pContext, FxU32 fpFlags)
{
#ifdef WINNT
    H3_MEMBASE0 *sstIOregs = (PH3_MEMBASE0) DFP_DEVINFOPTR(pContext)->MappedAddress[SST_IO_REGS_INDEX];
#else
    SstIORegs *sstIOregs = (void *)DFP_DEVINFOPTR(pContext)->RegBase[HWINFO_SST_IOREGS_INDEX];
#endif

    if (!fpFlags)    // If flags come in as 0; Turn off the display
    {
        panelOff(pContext); // Toggle 'SST_VIDEOIN_TVOUT_ENABLE' to off.
        return (0);
    }
    else if (fpFlags != 1)
        DFP_DEVINFOPTR(pContext)->dfpDisp.FpFlags = fpFlags;

#ifndef WINNT
	// tick reset for controller chip in 'vidSerialParallelPort' bit 31.
//***	sstIOregs->vidSerialParallelPort &= ~(1 << SST_SERPAR_TVOUTRESET_N);
	sstIOregs->vidSerialParallelPort &= ~SST_SERPAR_TVOUTRESET_N;
	DELAY(100);
//***	sstIOregs->vidSerialParallelPort |= (1 << SST_SERPAR_TVOUTRESET_N);
	sstIOregs->vidSerialParallelPort |= SST_SERPAR_TVOUTRESET_N;
#endif

    if (!flatPanelPresent (pContext))
        return (-1);

    panelOn( DFP_DEVINFOPTR(pContext) );

    // setup pixclk to pixdata delay
    sstIOregs->tmuGbeInit = (sstIOregs->tmuGbeInit & ~0xf8000) | 0;

    return (0);
}



#ifdef WINNT
int DFP_ProcessRequest(void * pContext, PVIDEO_REQUEST_PACKET RequestPacket)
{
    int status;
    ULONG dfpStatus;
    UCHAR ucTemp;

    if (sizeof(ULONG) >  RequestPacket->InputBufferLength)
    {
        VideoDebugPrint((0,"DFPPACKET input buffer too small\n"));
        return (0xffffffff);
    }

    status = 0xffffffff;

    switch (((PDFPPACKET)RequestPacket->InputBuffer)->dfpPacketFunc)
    {
    case dfpGetSetFlatPanelPhysical:
        if (RequestPacket->OutputBufferLength < (RequestPacket->StatusBlock->Information = sizeof(FlatPnlPhysical)))
        {
            VideoDebugPrint((0,"FlatPnlPhysical result buffer size mismatch\n"));
            break;
        }

        // copy input structure to output structure before calling flatPanelPhysical
        memcpy( RequestPacket->OutputBuffer, &(((PDFPPACKET)RequestPacket->InputBuffer)->dfpOptData), sizeof(FlatPnlPhysical));

        flatPanelPhysical (pContext, RequestPacket->OutputBuffer);
        status = 0;
        break;

    case dfpSetFlatPanelCtrl:
#if 0  //xlcd Not Ported From Win9x
        if(NAPALM_HW_DFP && IS_VOODOO3_X(pContext))
            mrtiSet (pContext, ((PDFPPACKET)RequestPacket->InputBuffer)->dfpOptData.dfpControl);
        else
            panelSet (pContext, ((PDFPPACKET)RequestPacket->InputBuffer)->dfpOptData.dfpControl);
#else
        panelSet (pContext, ((PDFPPACKET)RequestPacket->InputBuffer)->dfpOptData.dfpControl);
#endif
        status = 0;
        break;

    case dfpGetFlatPanelStatus:
        if (RequestPacket->OutputBufferLength < (RequestPacket->StatusBlock->Information = sizeof(DFPSTATUS)))
        {
            VideoDebugPrint((0,"dfpGetFlatPanelStatus result buffer size mismatch\n"));
            break;
        }

        ((PDFPSTATUS)RequestPacket->OutputBuffer)->panelPresent = flatPanelPresent(pContext);
        ((PDFPSTATUS)RequestPacket->OutputBuffer)->panelActiveAtBoot = DFP_DEVINFOPTR(pContext)->bDfpActiveAtBoot;
        ((PDFPSTATUS)RequestPacket->OutputBuffer)->panelActive = isPanelActive(pContext);
        status = 0;
        break;
    case dfpTurnOnFlatPanel:
        panelOn( pContext );
        status = 0;
        break;
    case dfpTurnOffFlatPanel:
        panelOff( pContext );
        status = 0;
        break;
    }

    return (status);

}

void DFP_Initialize(void * pContext)

{
    FlatPnlPhysical flatPnlDims;

#ifndef DANOTESTDFP
// No released Voodoo3 boards are supported by this driver, but there are a few prototype Voodoo3 DFP
// are used in testing.  These prototype boards can be enabled by defining DANOTESTDFP.
    if (IS_VOODOO3_X(pContext))
    {
        DFP_DEVINFOPTR(pContext)->bDfpActiveAtBoot  = FALSE;
        return;
    }
    else
#endif

    // if Bios says panel is active the first time we check then we believe it was present at boot.
    DFP_DEVINFOPTR(pContext)->bDfpActiveAtBoot  = (BOOLEAN)isPanelActive(pContext);

    // force dfpDisp.dispWidth, and dfpDisp.dispHeight to be filled in with values read from EDID
    flatPnlDims.fpFlags = FPFLAG_GET_DIMS;
	DFP_DEVINFOPTR(pContext)->dfpDisp.dispWidth = 0;     // insure it's zero so EDID is read.
    flatPanelPhysical(pContext, &flatPnlDims);

}

/*----------------------------------------------------------------------
Function name:  DFP_SetActiveState

Description:    Enable or Disable the DFP device as selected by value of alternate
                display device mask passed in.

Information:

Return:         FxU32     STB_FUNCTION_DFP if success,
                          0 if the current mode does not allow completion.
----------------------------------------------------------------------*/
FxU32 DFP_SetActiveState (PHW_DEVICE_EXTENSION HwDeviceExtension, FxU32 activeStateMask)
{
    PH3_MEMBASE0 RegisterMap = (PH3_MEMBASE0) HwDeviceExtension->MappedAddress[SST_IO_REGS_INDEX];
    ULONG xres = RegisterMap->vidScreenSize & 0xfff;
    ULONG result = STB_FUNCTION_DFP;


#if 0
    if (HwDeviceExtension->bDfpPresent)
    {
#endif
        // the card is DFP Capable, check if DFP is supposed to be on.
        if ((activeStateMask & STB_FUNCTION_DFP) == STB_FUNCTION_DFP)
        { // DFP should be turned on
            // check if DFP is off and needs to be turned on
            if (!isPanelActive(HwDeviceExtension))
            {
                // DFP is off and needs to be turned on, Check if the current mode is valid for DFP

                // Check for "Scaling DFP" or "Centering DFP".
                if (isPanelScaling( HwDeviceExtension ))
                {
                    // "Scaling DFP"

                    switch (xres)
                    {
                    // a complete list of mode that are legal for "Scaling DFP" ??? not final ???
                    case 640:
                    case 800:
                    case 1024:
                        // the current mode is valid for DFP turn it on.
                        panelOn( HwDeviceExtension );
                        break;

                    default:
                        // The current mode is not valid for DFP.
                        result = 0;
                    }
                }
                else
                {
                    // "Centering DFP"

                    // check if a DFP timing is available for this mode
                    if (HwDeviceExtension->ActiveFrequencyEntry->DfpData[0] == 0)
                    {
                        result = 0;
                    }
                    else
                    {
                        // the current mode is valid for DFP turn it on.
                        panelOn( HwDeviceExtension );
                    }

                }

                if (result == 0)
                    panelOn( HwDeviceExtension );   //try to turn it on anyway ???

            }

        }
        else
        {// DFP should be turned off
            // check if DFP is on and needs to be turned off
            if (isPanelActive(HwDeviceExtension))
                 panelOff( HwDeviceExtension );

        }
#if 0
    }
#endif
    return(result);
}
#else
/*----------------------------------------------------------------------
Function name:  panelGbl

Description:    Returns a ptr to the globaldata panel status dword

Information:    

Return:         INT     0 or -1.
----------------------------------------------------------------------*/
PDWORD panelGbl(void * pContext)
{
   if(DFP_DEVINFOPTR(pContext)->lpDriverData)
      return &(((GLOBALDATA *)DFP_DEVINFOPTR(pContext)->lpDriverData)->dwDFPState);
   else
      return NULL;
}

#endif //else def WINNT
//EOF