/* -*-c++-*- */
/* $Header: xlcd.c, 4, 9/22/99 11:05:15 AM, Ryan Bissell$ */
/*
** Copyright (c) 1995-1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished  -
** rights reserved under the Copyright Laws of the United States.
**
** File name:   xlcd.c
**
** Description: Support functions for xlcd.
**
** $Revision: 4$
** $Date: 9/22/99 11:05:15 AM$
**
** $History: xlcd.c $
** 
** *****************  Version 32  *****************
** User: Rbissell     Date: 9/07/99    Time: 2:16p
** Updated in $/devel/h5/Win9x/dx/minivdd
** Deployed the new modularized I2C code.
** 
** *****************  Version 31  *****************
** User: Dalev        Date: 8/20/99    Time: 4:20p
** Updated in $/devel/h5/Win9x/dx/minivdd
** Updated;  code now handles only the mrti xlcd chip.  Function names
** changed from Panel to mrti.
** 
** *****************  Version 30  *****************
** User: Rbissell     Date: 8/07/99    Time: 3:13p
** Updated in $/devel/h5/Win9x/dx/minivdd
** tvout merge from V3_OEM_100
**
** *****************  Version 29  *****************
** User: Andrew       Date: 7/16/99    Time: 2:06p
** Updated in $/devel/h5/Win9x/dx/minivdd
** Changed regBase and RegBase from single dword to array to support
** sparse register mapping
** 
** *****************  Version 28  *****************
** User: Cwilcox      Date: 7/08/99    Time: 1:11p
** Updated in $/devel/h5/Win9x/dx/minivdd
** Added runtime checking for Napalm versus Voodoo3.
** 
** *****************  Version 26  *****************
** User: Stb_rbissell Date: 5/14/99    Time: 9:36p
** Updated in $/devel/h3/win95/dx/minivdd
** tvout and dfp changes
** 
** *****************  Version 25  *****************
** User: Stuartb      Date: 3/18/99    Time: 11:48a
** Updated in $/devel/h3/Win95/dx/minivdd
** Added the function panelFixupVGA to set vertical total CRTC register
** when transitioning to a full screen DOS box.
** 
** *****************  Version 24  *****************
** User: Stuartb      Date: 3/18/99    Time: 10:32a
** Updated in $/devel/h3/Win95/dx/minivdd
** Add hack for 1280 in 1280 out.  Tweaked timings.  Add DDC select for
** panel.  Added comments.
** 
** *****************  Version 22  *****************
** User: Stuartb      Date: 2/18/99    Time: 2:57p
** Updated in $/devel/h3/Win95/dx/minivdd
** Added #idef'd hacks for testing Hitachi panel.  Added filter
** coefficient setting code.  Fixed refresh rate bug.
** 
** *****************  Version 21  *****************
** User: Stuartb      Date: 2/16/99    Time: 5:11p
** Updated in $/devel/h3/Win95/dx/minivdd
** Changed 8x6 capture clock due to modetabl changes.  Added more support
** for 12x10 panels.
** 
** *****************  Version 20  *****************
** User: Stuartb      Date: 2/08/99    Time: 3:39p
** Updated in $/devel/h3/Win95/dx/minivdd
** Hack to allow (VGA std) Princeton and Caompaq FP500 panels to work.
** Gateway panels may no long work.
** 
** *****************  Version 19  *****************
** User: Stuartb      Date: 2/08/99    Time: 8:58a
** Updated in $/devel/h3/Win95/dx/minivdd
** Changes to fix lcd boot and simultaneous VMI & TV/LCD.
** 
** *****************  Version 18  *****************
** User: Stuartb      Date: 2/02/99    Time: 5:02p
** Updated in $/devel/h3/Win95/dx/minivdd
** Invert DCLK (PANEL_CTRL reg).  Makes noise at 10x7 go away.
** 
** *****************  Version 17  *****************
** User: Stuartb      Date: 2/02/99    Time: 10:59a
** Updated in $/devel/h3/Win95/dx/minivdd
** Added code to support turning LCD on/off on sleep.
** 
** *****************  Version 16  *****************
** User: Agus         Date: 2/01/99    Time: 4:36p
** Updated in $/devel/h3/Win95/dx/minivdd
** Move IS_H4 macro define to h3vdd.h
** 
** *****************  Version 15  *****************
** User: Stuartb      Date: 1/28/99    Time: 4:35p
** Updated in $/devel/h3/Win95/dx/minivdd
** Eliminate negative coeffs in filter due to chip problems.
** 
** *****************  Version 14  *****************
** User: Stuartb      Date: 1/26/99    Time: 5:14p
** Updated in $/devel/h3/Win95/dx/minivdd
** Code to enable interpolation filter.  Minor changes for -04 chip.
** Actual DCLK is now in Xcap.
** 
** *****************  Version 13  *****************
** User: Stuartb      Date: 1/22/99    Time: 5:14p
** Updated in $/devel/h3/Win95/dx/minivdd
** Tweaked parameters, fixed bug in pllCalc, added code for 12x10 panels.
** 
** *****************  Version 12  *****************
** User: Stuartb      Date: 1/21/99    Time: 11:18a
** Updated in $/devel/h3/Win95/dx/minivdd
** Allow panels greater that 1024x768.
** 
** *****************  Version 11  *****************
** User: Michael      Date: 1/15/99    Time: 9:59a
** Updated in $/devel/h3/Win95/dx/minivdd
** Implement the 3Dfx/STB unified header.
** 
*/


#include "h3vdd.h"
#include "h3.h"
#define VDDONLY
#include "devtable.h"
#include "h3g.h"
#undef  VDDONLY
#ifdef OLDI2C
  #include "i2c.h"
#else
  #include "i2c/di_i2c.h"
  #define DELAY(usecs) CM_Yield(usecs, CM_YIELD_RESUME_EXEC)
#endif

//extern const struct i2cmask I2C_PROTO;
//extern const DWORD I2C_INITVAL;
//extern int read_sda (PI2CMASK);
//extern int WaitTime (FxU32);


#include "xlcd.h"
#include "dfp.h"
#include "dfpdisp.h"

#if 0    // Arghhh, they changed the hardware again
#undef IS_3DFX_REF_BOARD
#define IS_3DFX_REF_BOARD(pDev)  ((pDev->dwSubSystemID & 0xff0000) < 0x00300000)


const DWORD DDC_CRT_INITVAL = (DWORD)((1 << DDC_ENABLE)      |
									  (1 << DDC_SCL_OUT_BIT) |
									  (1 << DDC_SDA_OUT_BIT) |
									  (1 << TVOUT_RST_BIT));
const DWORD DDC_LCD_INITVAL = (DWORD)((1 << DDC_ENABLE)      |
									  (1 << DDC_SCL_OUT_BIT) |
									  (1 << DDC_SDA_OUT_BIT) |
									  (SST_SERPAR_GPIO_1)    |
									  (1 << TVOUT_RST_BIT));
const struct i2cmask DDC_PROTO = {0, 0, 0, 1, 0, DDC_ENABLE, DDC_SCL_OUT_BIT, \
						DDC_SDA_OUT_BIT, DDC_SCL_IN_BIT, DDC_SDA_IN_BIT};
#endif

FxU8 LCD_DEBUG[64];

#define XLCD_ADDR     0xe0
#define INTERPOLATE   1       // off til we get things going
#define RETRIES       4

#if XLCD_DEBUG
extern char *XlcdRegNames[];
#endif

#define RGB888_BPP    3
#define SYNC_INV      (CAP_FLAG_INV_HSYNC | CAP_FLAG_INV_VSYNC)

static const DWORD Xlcd_DclkMatrix[][5] = {
/*                             6x4     8x6     10x7  12x10   */
/* for 1024x768 panels  */  { 68011,  64982,  65455,      0},
/* for 1280x1024 panels */  {111682, 109772, 109176, 108000},
};

const XLCD_CAPTURE Xcap_1280 = {1280, 238, 1024, 37, 0, 108000};
const XLCD_CAPTURE Xcap_1024 = {1024, 144, 768, 29, SYNC_INV, 65455};
const XLCD_CAPTURE Xcap_800 =  {800, 72, 600, 23, 0, 65455};
const XLCD_CAPTURE Xcap_640 =  {640, 41, 480, 31, SYNC_INV, 68011};


const FxU8 InterpCoeffs[][64] = {
{	0,   1, 127,  0,    // bilinear
	0,   8, 120,  0,
	0,  16, 112,  0,
	0,  24, 104,  0,
	0,  32,  96,  0,
	0,  40,  88,  0,
	0,  48,  80,  0,
	0,  56,  72,  0,
	0,  64,  64,  0,
	0,  72,  56,  0,
	0,  80,  48,  0,
	0,  88,  40,  0,
	0,  96,  32,  0,
	0, 104,  24,  0,
	0, 112,  16,  0,
	0, 120,   8,  0},

{   0,   1, 127, 0,     // gausian
	0,   8, 120, 0,
	2,  14, 110, 2,
	3,  21, 101, 3,
	4,  28,  92, 4,
	4,  36,  84, 4,
	6,  42,  74, 6,
	7,  49,  65, 7,
	8,  56,  56, 8,
	7,  65,  49, 7,
	6,  74,  42, 6,
	4,  84,  36, 4,
	4,  92,  28, 4,
	3, 101,  21, 3,
	2, 110,  14, 2,
	0, 120,   8, 0},


{0x00,   1, 127,  0x00,    // sharp

 0x00,   8, 120,  0x00,
//// 0x81,   8, 122,  0x81,

// 0x81,   8, 120,  0x81,
 0x82,  18, 114,  0x82,
 0x84,  28, 108,  0x84,
 0x84,  36, 100,  0x84,
 0x84,  44,  92,  0x84,
 0x86,  54,  86,  0x86,
 0x88,  64,  80,  0x88,
 0x88,  72,  72,  0x88,
 0x88,  80,  64,  0x88,
 0x86,  86,  54,  0x86,
 0x84,  92,  44,  0x84,
 0x84, 100,  36,  0x84,
 0x84, 108,  28,  0x84,
 0x82, 114,  18,  0x82,
 0x00, 120,   8,  0x00},

/*
{	0x00,   1, 127,  0x00,   // 0x80,	1, 127,  0x80,   // sharp
	0x81,   8, 122,  0x81,   // 0x81,   8, 120,  0x81,
	0x82,  18, 114,  0x82,
	0x83,  27, 107,  0x83,
	0x84,  36, 100,  0x84,
	0x84,  44,  92,  0x84,
	0x86,  54,  86,  0x86,
	0x87,  63,  79,  0x87,
	0x88,  72,  72,  0x88,
	0x87,  79,  63,  0x87,
	0x86,  86,  54,  0x86,
	0x84,  92,  44,  0x84,
	0x84, 100,  36,  0x84,
	0x83, 107,  27,  0x83,
	0x82, 114,  18,  0x82,
	0x00, 120,   8,  0x00},   // 0x80, 120,   8,  0x80},
*/


{	0x00,  1, 127, 0x00,      // extra sharp (what stu called "crinkly"?)
	0x82, 10, 122, 0x82, 
	0x84, 20, 116, 0x84, 
	0x86, 30, 110, 0x86, 
	0x88, 40, 104, 0x88, 
	0x8a, 50,  98, 0x8a, 
	0x8c, 60,  92, 0x8c, 
	0x8e, 70,  86, 0x8e, 
	0x90, 80,  80, 0x90, 
	0x8e, 86,  70, 0x8e, 
	0x8c, 92,  60, 0x8c, 
	0x8a, 98,  50, 0x8a, 
	0x88, 104, 40, 0x88, 
	0x86, 110, 30, 0x86, 
	0x84, 116, 20, 0x84, 
	0x82, 122, 10, 0x82},

/*
{   0x00,  1, 127, 0x00,      // crinkly
	0x82, 10, 122, 0x82,
	0x84, 20, 116, 0x84,
	0x86, 30, 110, 0x86,
	0x88, 40, 104, 0x88,
	0x8a, 50,  98, 0x8a,
	0x8c, 60,  92, 0x8c,
	0x8e, 70,  86, 0x8e,
	0x90, 80,  80, 0x90,
	0x8e, 86,  70, 0x8e,
	0x8c, 92,  60, 0x8c,
	0x8a, 98,  50, 0x8a,
	0x88, 104, 40, 0x88,
	0x86, 110, 30, 0x86,
	0x84, 116, 20, 0x84,
	0x82, 122, 10, 0x82}


*/

};

#undef PNL_1280_TEST   // to test 12x10 panel with no EDID, best we can do

/*----------------------------------------------------------------------
Function name:  mrtiSendByte

Description:    Send a byte to the XLCD device.

Information:    

Return:         INT     0 or -1.
----------------------------------------------------------------------*/
int mrtiSendByte(PDEVTABLE pDev, I2CKEY key, BYTE data)
{
	int	i;

	for (i = 0; i < 8; i++)
	{
	  i2c_setsda(pDev, key, (data << i) & 0x80, 0);
	  i2c_setscl(pDev, key, 1, 0);
	  i2c_setscl(pDev, key, 0, 0);
	}
	i2c_setscl(pDev, key, 1, 0);// these two lines are backwards, but then again so
	i2c_setsda(pDev, key, 1, 0);// is MRTI chip
	i = (i2c_getsda(pDev, key) == 1) ? -1 : 0;
	i2c_setscl(pDev, key, 0, 0);
	return (i);
}


/*----------------------------------------------------------------------
Function name:  xlcdSendByte

Description:    Send a byte to the XLCD device.

Information:    

Return:         INT     0 or -1.
----------------------------------------------------------------------*/
int xlcdSendByte(PDEVTABLE pDev, I2CKEY key, BYTE data)
{
	int	i;

	for (i = 0; i < 8; i++)
	{
	  i2c_setsda(pDev, key, (data << i) & 0x80, 0);
	  i2c_setscl(pDev, key, 1, 0);
	  i2c_setscl(pDev, key, 0, 0);
	}
	i2c_setsda(pDev, key, 1, 0);
	i2c_setscl(pDev, key, 1, 0);
	i = (i2c_getsda(pDev, key) == 1) ? -1 : 0;
	i2c_setscl(pDev, key, 0, 0);
	return (i);
}


/*----------------------------------------------------------------------
Function name:  writemrtiReg

Description:    

Information:    

Return:         INT     0 or -1.
----------------------------------------------------------------------*/
int writemrtiReg(PDEVTABLE pDev, I2CKEY key, int reg, int val)
{
	int retries;

	LCD_DEBUG[reg & 63] = val;
	for (retries = RETRIES + 1; retries--;   )
	{
		i2c_start(pDev, key);    // Send '1,0'
		if (!mrtiSendByte(pDev, key, XLCD_ADDR))  // Address 0xE0.
			if (!mrtiSendByte(pDev, key, 0x00))
				if (!mrtiSendByte(pDev, key, (char)reg))  // Send register adrs.
					if (!mrtiSendByte(pDev, key, (char)val))  // Send register contents.
						break;
		i2c_stop(pDev, key);     // Send '0,1'.  Shouldn't come here unless there's a problem!
	}                    
	i2c_stop(pDev, key);       // Send '0,1'.  This is where we should go.
	if (retries != RETRIES)
		Debug_Printf ("%d write retries\n", RETRIES - retries);

#if 0
	printf ("(%02x) %s written %02x, read %02x\n", reg, XlcdRegNames[reg],
						val, dbgReadXLCDReg (reg, XLCD_ADDR));
#endif
// If failed on ALL retries; return -1 (fail), else return 0 (pass).
	return (retries <= 0 ? -1 : 0);
}


/*----------------------------------------------------------------------
Function name:  xlcdFilter

Description:    

Information:    

Return:         INT     0 or -1.
----------------------------------------------------------------------*/
int xlcdFilter(PDEVTABLE pDev, I2CKEY key)
{
	int retries;
	int addr = 0;
	FxU8 const *coeffs;

	if (pDev->dfpDisp.FpFlags & FPFLAG_FILTER_HARD)
		coeffs = InterpCoeffs[!(pDev->dfpDisp.FpFlags & FPFLAG_FILTER_SOFT) ? 2 : 3];
	else
		coeffs = InterpCoeffs[!(pDev->dfpDisp.FpFlags & FPFLAG_FILTER_SOFT) ? 0 : 1];

	for (retries = RETRIES + 1; retries--;   )
	{
		i2c_start(pDev, key);
		if (!mrtiSendByte(pDev, key, XLCD_ADDR))
		{
			if (!mrtiSendByte(pDev, key, 0x60))
			{
				if (!mrtiSendByte(pDev, key, (FxU8)addr))  // send starting addr
				{
					// setup H & V filters identically
					for (     ; addr < 128; addr++)
					{
						if (mrtiSendByte(pDev, key, coeffs[addr & 63]))
							break;
					}
				}
			}
		}
		i2c_stop(pDev, key);
		if (addr == 128)
			break;
		if (addr & 1)   // need to always send pairs of bytes
			addr--;
	}
	if (retries != RETRIES)
		Debug_Printf ("%d h filter retries\n", RETRIES - retries);

	return (retries <= 0 ? -1 : 0);
}


/*----------------------------------------------------------------------
Function name:  PLLcalc

Description:    Determine PLL m/n that yields the Dclk closest to that
                requested.

Information:    

Return:         INT
----------------------------------------------------------------------*/
int PLLcalc(int refFreq, int desiredFreq)
{
	long freq, diff, closest;
	int retval, m, n;

	closest = 0x7fffffff;
	for (n = 3; n < 18; n++)
	{
		for (m = 3; m < 66; m++)
		{
			freq = (long)(refFreq * m) / n;
			if ((diff = freq - desiredFreq) < 0)
				diff = 0 - diff;
			if (diff < closest)
			{
				closest = diff;
				retval = (m << 8) | n;
			}
		}
	}
	retval -= 0x0202;    // as we program m-2 and n-2
	return (retval);
}


/*----------------------------------------------------------------------
Function name:  mrtiOff

Description:    Turn off the mrti device.

Information:    

Return:         VOID
----------------------------------------------------------------------*/
void mrtiOff(PDEVTABLE pDev, I2CKEY key)
{
  writemrtiReg(pDev, key, PANEL_POWER, 0);

  /********
  struct i2cmask i2cMask = I2C_PROTO;
	SstIORegs *sstIOregs = (void *)pDev->RegBase[HWINFO_SST_IOREGS_INDEX];
  
	i2cMask.pReg = (DWORD *)&sstIOregs->vidSerialParallelPort;
   *i2cMask.pReg = ((*i2cMask.pReg & ~I2C_INITMASK) | I2C_INITVAL | 0x80000000);
	writemrtiReg (&i2cMask, PANEL_POWER, 0);
  *********/
}


/*----------------------------------------------------------------------
Function name:  mrtiOn

Description:    Turn on the mrti device.

Information:    

Return:         VOID
----------------------------------------------------------------------*/
void mrtiOn(PDEVTABLE pDev, I2CKEY key)
{
  writemrtiReg(pDev, key, PANEL_POWER, 0x3D);

  /********
	struct i2cmask i2cMask = I2C_PROTO;
	SstIORegs *sstIOregs = (void *)pDev->RegBase[HWINFO_SST_IOREGS_INDEX];

	i2cMask.pReg = (DWORD *)&sstIOregs->vidSerialParallelPort;
   *i2cMask.pReg = ((*i2cMask.pReg & ~I2C_INITMASK) | I2C_INITVAL | 0x80000000);
	writemrtiReg (&i2cMask, PANEL_POWER, 0x3d);
  *********/
}


/*----------------------------------------------------------------------
Function name:  mrtiConfig

Description:    Configure the panel device.

Information:    

Return:         INT     0 is always returned.
----------------------------------------------------------------------*/
int mrtiConfig(PDEVTABLE pDev, XLCD_CAPTURE *cap, DFP_DISPLAY *disp, int pd)
{
	SstIORegs *sstIO = (void *)pDev->RegBase[HWINFO_SST_IOREGS_INDEX];
	int v, start, m, n;
	int hzoom = (long)(cap->capWidth * 2048L) / disp->dispWidth;
	int vzoom = (long)(cap->capHeight * 2048L) / disp->dispHeight;
	//*******struct i2cmask ddcMask = I2C_PROTO;
  I2CKEY key;

  key = i2c_getaccess(pDev, I2C_LCDSCALER, I2C_NORMALSPEED);
  if (key == I2C_NOTAKEY)
    return (0);


   panelOn( pDev );     // Set 'vidInFormat' bit 15 to a 1.

// Set MRTI chip to sample 12-bit data at falling-then-rising edges.
	sstIO->vidInFormat &= ~(SST_VIDEOIN_G4_FOR_POSEDGE);

	// setup pixclk to pixdata delay
	sstIO->tmuGbeInit = (sstIO->tmuGbeInit & ~0xf8000) | 0;

  /********
	ddcMask.pReg = (DWORD *)&sstIO->vidSerialParallelPort;
  *ddcMask.pReg = ((*ddcMask.pReg & ~I2C_INITMASK) | I2C_INITVAL | 0x80000000);
  *********/

#if 0
	printf ("capW = %d, capH = %d, capHofs = %d, capVofs = %d\n",
				cap->capWidth, cap->capHeight, cap->capHofst, cap->capVofst);
	printf ("dispWidth = %d, dispHeight = %d, dispHtotal = %d, dispVtotal = %d, dispHblk = %d, dispVBlk = %d\n", disp->dispWidth, disp->dispHeight,
	                             disp->htotal, disp->vtotal, disp->hblank, 
								 disp->vblank);
#endif

	if (pd)        // Turn DFP off before programming MRTI chip.
		mrtiOff(pDev, key);
	
	// Calulate PLL m/n
	n = PLLcalc (14318, cap->clkFreq);
	m = n >> 8;

	/*
	PLL_CONTROL = 0x0e = (M - 2) = 0x39
	MISC_GP_REG = 0x0f = (N - 2) = 0x0b   , need docs
	64.98mc = (14.31818 * 59) / 13
	*/
	writemrtiReg (pDev, key, PLL_CONTROL, m);
	writemrtiReg (pDev, key, MISC_GP_REG, 0x30 | (n & 15));

	// Display parameters
	writemrtiReg (pDev, key, DISPLAY_HTOTAL, disp->htotal / 8);
	writemrtiReg (pDev, key, DISPLAY_HSTOP, (disp->dispWidth / 8) - 1);
	writemrtiReg (pDev, key, DISPLAY_HBLK_START,
						   (disp->dispWidth + disp->hsyncOfst) / 8);
	writemrtiReg (pDev, key, DISPLAY_VTOTAL_LO, disp->vtotal & 0xff);
	writemrtiReg (pDev, key, DISPLAY_VTOTAL_HI, disp->vtotal >> 8);
	disp->dispHeight--;
	writemrtiReg (pDev, key, DISPLAY_VSTOP_LO, disp->dispHeight & 0xff);
	writemrtiReg (pDev, key, DISPLAY_VSTOP_HI, disp->dispHeight >> 8);
	disp->dispHeight++;
	start = disp->dispHeight + disp->vsyncOfst;
	writemrtiReg (pDev, key, DISPLAY_VBLK_START_LO, start & 255);
	writemrtiReg (pDev, key, DISPLAY_VBLK_START_HI, start >> 8);
	writemrtiReg (pDev, key, DISPLAY_VSYNC_WIDTH, disp->vsyncWidth);
	writemrtiReg (pDev, key, DISPLAY_HSYNC_WIDTH, disp->hsyncWidth / 8);

	// Capture parameters
	writemrtiReg (pDev, key, CAPTURE_HSTART, cap->capHofst >> 3);
	writemrtiReg (pDev, key, CAPTURE_PEL_ALIGNMENT, cap->capHofst & 7);
	writemrtiReg (pDev, key, CAPTURE_HSTOP, (cap->capWidth / 8) + (cap->capHofst >> 3));
	writemrtiReg (pDev, key, CAPTURE_VSTART_LO, cap->capVofst);
	writemrtiReg (pDev, key, CAPTURE_VSTART_HI, 0);
	cap->capHeight += (cap->capVofst + 1);
	writemrtiReg (pDev, key, CAPTURE_VSTOP_LO, cap->capHeight & 0xff);
	writemrtiReg (pDev, key, CAPTURE_VSTOP_HI, cap->capHeight >> 8);
	cap->capHeight -= (cap->capVofst + 1);

	// Zoom parameters
	writemrtiReg (pDev, key, DISPLAY_HZOOM_LO, hzoom & 0xff);
	writemrtiReg (pDev, key, DISPLAY_VZOOM_LO, vzoom & 0xff);
	writemrtiReg (pDev, key, DISPLAY_HV_ZOOM_HI, ((hzoom >> 8) << 4) | (vzoom >> 8));
	v = ((cap->capWidth * RGB888_BPP) / 8) - 1;
	writemrtiReg (pDev, key, CAPTURE_LINE_WIDTH_LO, v & 255);
	writemrtiReg (pDev, key, CAPTURE_LINE_WIDTH_HI, v >> 8);

	// Miscellaneous parameters
	writemrtiReg (pDev, key, INTERRUPT_MASK, 0);

	writemrtiReg (pDev, key, CAPTURE_IP_CTRL, cap->flags & 6);

	/*
	HACK ALERT!!!!!!!!!!!!!!

	The only way to get a stable 1:1 picture on the Hitachi 1280x1024 panel 
	is to frig the chip in test mode and flip Dclk phase.  This is all wrong
	but it works.  It won't work when scaling.  The Htotal was changed slightly
	from 1688 (VESA) to 1680 to make up for the fact the clock is wrong.  This
	hack seems to stem from the phase relationship of DENA vs Hsync which 
	changes in this mode.  Assume v = 0 for non-hack case.    3/17/99 SMB
	*/

	v = (disp->dispWidth == 1280 && cap->capWidth == 1280);
	writemrtiReg (pDev, key, PANEL_CTRL, v ? 0x18 : 0x30);
	writemrtiReg (pDev, key, TEST_MODE0, v);
	writemrtiReg (pDev, key, PANEL_DITHER, 0);
	writemrtiReg (pDev, key, TEST_MODE1, 0);
	writemrtiReg (pDev, key, ENGINE_ENABLES, 0);

#if INTERPOLATE
	writemrtiReg (pDev, key, DISPLAY_CTRL, (hzoom == 2048 ? 0 : 0x05) |
	                                     (vzoom == 2048 ? 0 : 0x0a));
#else
	writemrtiReg (pDev, key, DISPLAY_CTRL, (hzoom == 2048 ? 0 : 0x01) |
	                                     (vzoom == 2048 ? 0 : 0x02));
#endif

// Filter coeffs - only use if DFP resolution does NOT match
// desktop resolution.
	if (INTERPOLATE && (hzoom != 2048 || vzoom != 2048))
   {
      // we're thinking we always want to use this filter - even if
      // DFP resolution MATCHES desktop resolution.
      pDev->dfpDisp.FpFlags = (pDev->dfpDisp.FpFlags &
                               ~(FPFLAG_FILTER_HARD | FPFLAG_FILTER_SOFT)) | FPFLAG_FILTER_HARD;
		  xlcdFilter (pDev, key);
   }

	writemrtiReg (pDev, key, CAPTURE_ENABLE, 1);
	mrtiOn(pDev, key);

  i2c_endaccess(pDev, key);
	return (0);
}

/*----------------------------------------------------------------------
Function name:  mrtiPresent

Description:    Query for a panel device.

Information:    

Return:         INT     0 for no panel device, else panel device
----------------------------------------------------------------------*/
int mrtiPresent(PDEVTABLE pDev)
{
/********
	struct i2cmask ddcMask;
	SstIORegs *sstIO = (void *)pDev->RegBase[HWINFO_SST_IOREGS_INDEX];
  DWORD vidSerialParallel = sstIO->vidSerialParallelPort;
*********/
  I2CKEY key;
	int status = 0;

  key = i2c_getaccess(pDev, I2C_LCDSCALER, I2C_NORMALSPEED);
  if (key == I2C_NOTAKEY)
    return 0;

  panelOn(pDev);
  status = writemrtiReg(pDev, key, TEST_MODE1, 0);
  panelOff(pDev);
  i2c_endaccess(pDev, key);

  return (!status);

/********
   ddcMask = I2C_PROTO;
	ddcMask.pReg = (DWORD *)&sstIO->vidSerialParallelPort;
// Set bits 23,24,25 & 31 in sstIO->vidSerialParallelPort.
// Leave GPIO_1 bit OFF for access to MRTI ASIC.
   *ddcMask.pReg = ((*ddcMask.pReg & ~I2C_INITMASK) | I2C_INITVAL);
// Set bit 15 to on in 'vidInFormat' register.
   panelOn( pDev );
	status = writemrtiReg (&ddcMask, TEST_MODE1, 0);
   panelOff( pDev );
// restore orig value
	sstIO->vidSerialParallelPort = vidSerialParallel;
	return (!status);
*********/
}

/*----------------------------------------------------------------------
Function name:  mrtiSet

Description:    Progam's the MRTI chip with the correct scaling
                parameters.
Information:    

Return:         INT     0 or -1.
----------------------------------------------------------------------*/
int mrtiSet(PDEVTABLE pDev, DWORD fpFlags)
{
  I2CKEY key;
	SstIORegs *sstIOregs = (void *)pDev->RegBase[HWINFO_SST_IOREGS_INDEX];

// Get the CRT's current Vertical & Horizontal resolutions.
  DWORD yres = (sstIOregs->vidScreenSize >> 12) & 0xfff;
  DWORD xres = sstIOregs->vidScreenSize & 0xfff;
	XLCD_CAPTURE xcap = Xcap_1024;
#ifndef PNL_1280_TEST
	DFP_DISPLAY disp = DFPdisp_1024;
#else
	DFP_DISPLAY disp = DFPdisp_1280;
#endif

  key = i2c_getaccess(pDev, I2C_LCDSCALER, I2C_NORMALSPEED);
  if (key == I2C_NOTAKEY)
    return (-1);


  if (!fpFlags)    // If flags come in as 0; Turn off the display
	{
		// disable BIOS support through SCRATCH_REG2 ( can we put this in panelOff() )?
		outp ((FxU16)(pDev->IoBase + 0xd4), 0x1e);
		outp ((FxU16)(pDev->IoBase + 0xd5), (FxU8)(inp ((FxU16)(pDev->IoBase + 0xd5)) & ~4));
		mrtiOff(pDev, key);
    panelOff(pDev); // Toggle 'SST_VIDEOIN_TVOUT_ENABLE' to off.
    i2c_endaccess(pDev, key);
		return (0);
	}
	else if (fpFlags != 1)
		pDev->dfpDisp.FpFlags = fpFlags;

  i2c_endaccess(pDev, key);

	// tick reset for controller chip in 'vidSerialParallelPort' bit 31.
	sstIOregs->vidSerialParallelPort &= ~(1 << SST_SERPAR_TVOUTRESET_N);
	DELAY(100);
	sstIOregs->vidSerialParallelPort |= (1 << SST_SERPAR_TVOUTRESET_N);

	if (!flatPanelPresent (pDev))
		return (-1);

// Now get the EDID from the Panel.
	panelDDC (pDev, &disp);
	switch (disp.dispWidth)
	{
		case 1280:    // do it again with proper defaults
			disp = DFPdisp_1280;
			panelDDC (pDev, &disp);
			break;
		default:
			break;
	}

	// Set 'xcap' with the parameters to stretch the DFP to
   // the resolution of the desktop.  Note we assume only a 
   // 1024 (or 1280) DFP is connected.
	switch (xres)
	{
		case 640:
			xcap = Xcap_640;
			if (disp.dispWidth == 1280)
			{
				disp.htotal -= 32;     // ???????????
				xcap.clkFreq = Xlcd_DclkMatrix[1][0];
			}
			break;
		case 800:
			xcap = Xcap_800;
			if (disp.dispWidth == 1280)
				xcap.clkFreq = Xlcd_DclkMatrix[1][1];
			break;
		case 1024:
			if (disp.dispWidth == 1280)
				xcap.clkFreq = Xlcd_DclkMatrix[1][2];
			break;
		case 1280:
			xcap = Xcap_1280;
			break;
	}

	mrtiConfig(pDev, &xcap, &disp, 1);

	// enable BIOS support through SCRATCH_REG2, 3d4.1e.
	outp ((FxU16)(pDev->IoBase + 0xd4), 0x1e);
	outp ((FxU16)(pDev->IoBase + 0xd5),
		(FxU8)(inp ((FxU16)(pDev->IoBase + 0xd5)) | 4));
	return (0);
}